from __future__ import annotations

import datetime

import daft
from daft.sql.sql import SQLCatalog


def test_temporals():
    df = daft.from_pydict(
        {
            "times": [
                datetime.time(1, 2, 3, 4),
                datetime.time(2, 3, 4, 5),
                datetime.time(3, 4, 5, 6),
                datetime.time(4, 5, 6, 7),
                datetime.time(5, 6, 7, 8),
                None,
            ],
            "datetimes": [
                datetime.datetime(2021, 1, 1, 23, 59, 58, 999_999),
                datetime.datetime(2021, 1, 2, 0, 0, 0, 0),
                datetime.datetime(2021, 1, 2, 1, 2, 3, 500_000),
                datetime.datetime(2021, 2, 2, 1, 2, 3, 100_000),
                datetime.datetime(1999, 1, 31, 1, 1, 1, 50),
                None,
            ],
        }
    )
    catalog = SQLCatalog({"test": df})

    expected = df.select(
        daft.col("datetimes").dt.date().alias("date"),
        daft.col("datetimes").dt.day().alias("day"),
        daft.col("datetimes").dt.day_of_week().alias("day_of_week"),
        daft.col("datetimes").dt.day_of_month().alias("day_of_month"),
        daft.col("datetimes").dt.day_of_year().alias("day_of_year"),
        daft.col("datetimes").dt.week_of_year().alias("week_of_year"),
        daft.col("datetimes").dt.hour().alias("hour"),
        daft.col("datetimes").dt.minute().alias("minute"),
        daft.col("datetimes").dt.month().alias("month"),
        daft.col("datetimes").dt.second().alias("second"),
        daft.col("datetimes").dt.millisecond().alias("millisecond"),
        daft.col("datetimes").dt.microsecond().alias("microsecond"),
        daft.col("datetimes").dt.nanosecond().alias("nanosecond"),
        daft.col("datetimes").dt.quarter().alias("quarter"),
        daft.col("datetimes").dt.year().alias("year"),
        daft.col("datetimes").dt.unix_date().alias("unix_date"),
        daft.col("datetimes").dt.to_unix_epoch().alias("to_unix_epoch"),
        daft.col("datetimes").dt.to_unix_epoch("s").alias("to_unix_epoch_s"),
        daft.col("datetimes").dt.to_unix_epoch("ms").alias("to_unix_epoch_ms"),
        daft.col("datetimes").dt.to_unix_epoch("us").alias("to_unix_epoch_us"),
        daft.col("datetimes").dt.to_unix_epoch("ns").alias("to_unix_epoch_ns"),
        daft.col("datetimes").dt.to_unix_epoch("seconds").alias("to_unix_epoch_seconds"),
        daft.col("datetimes").dt.to_unix_epoch("milliseconds").alias("to_unix_epoch_milliseconds"),
        daft.col("datetimes").dt.to_unix_epoch("microseconds").alias("to_unix_epoch_microseconds"),
        daft.col("datetimes").dt.to_unix_epoch("nanoseconds").alias("to_unix_epoch_nanoseconds"),
        daft.col("datetimes").dt.strftime().alias("date_str"),
        daft.col("times").dt.strftime().alias("time_str"),
    ).collect()

    actual = daft.sql(
        """
    SELECT
        date(datetimes) as date,
        day(datetimes) as day,
        day_of_week(datetimes) as day_of_week,
        day_of_month(datetimes) as day_of_month,
        day_of_year(datetimes) as day_of_year,
        week_of_year(datetimes) as week_of_year,
        hour(datetimes) as hour,
        minute(datetimes) as minute,
        month(datetimes) as month,
        second(datetimes) as second,
        millisecond(datetimes) as millisecond,
        microsecond(datetimes) as microsecond,
        nanosecond(datetimes) as nanosecond,
        quarter(datetimes) as quarter,
        year(datetimes) as year,
        unix_date(datetimes) as unix_date,
        to_unix_epoch(datetimes) as to_unix_epoch,
        to_unix_epoch(datetimes, 's') as to_unix_epoch_s,
        to_unix_epoch(datetimes, 'ms') as to_unix_epoch_ms,
        to_unix_epoch(datetimes, 'us') as to_unix_epoch_us,
        to_unix_epoch(datetimes, 'ns') as to_unix_epoch_ns,
        to_unix_epoch(datetimes, 's') as to_unix_epoch_seconds,
        to_unix_epoch(datetimes, 'ms') as to_unix_epoch_milliseconds,
        to_unix_epoch(datetimes, 'us') as to_unix_epoch_microseconds,
        to_unix_epoch(datetimes, 'ns') as to_unix_epoch_nanoseconds,
        strftime(datetimes) as date_str,
        strftime(times) as time_str,
    FROM test
    """,
        catalog=catalog,
    ).collect()

    assert actual.to_pydict() == expected.to_pydict()


def test_extract():
    df = daft.from_pydict(
        {
            "datetimes": [
                datetime.datetime(2021, 1, 1, 23, 59, 58, 999_999),
                datetime.datetime(2021, 1, 2, 0, 0, 0, 0),
                datetime.datetime(2021, 1, 2, 1, 2, 3, 500_000),
                datetime.datetime(2021, 2, 2, 1, 2, 3, 100_000),
                datetime.datetime(1999, 1, 31, 1, 1, 1, 50),
                None,
            ]
        }
    )

    expected = df.select(
        daft.col("datetimes").dt.date().alias("date"),
        daft.col("datetimes").dt.day().alias("day"),
        daft.col("datetimes").dt.day_of_week().alias("day_of_week"),
        daft.col("datetimes").dt.day_of_month().alias("day_of_month"),
        daft.col("datetimes").dt.day_of_year().alias("day_of_year"),
        daft.col("datetimes").dt.week_of_year().alias("week_of_year"),
        daft.col("datetimes").dt.hour().alias("hour"),
        daft.col("datetimes").dt.minute().alias("minute"),
        daft.col("datetimes").dt.month().alias("month"),
        daft.col("datetimes").dt.second().alias("second"),
        daft.col("datetimes").dt.millisecond().alias("millisecond"),
        daft.col("datetimes").dt.microsecond().alias("microsecond"),
        daft.col("datetimes").dt.nanosecond().alias("nanosecond"),
        daft.col("datetimes").dt.quarter().alias("quarter"),
        daft.col("datetimes").dt.year().alias("year"),
        daft.col("datetimes").dt.unix_date().alias("unix_date"),
    ).collect()

    actual = daft.sql("""
    SELECT
        extract(date from datetimes) as date,
        extract(day from datetimes) as day,
        extract(day_of_week from datetimes) as day_of_week,
        extract(day_of_month from datetimes) as day_of_month,
        extract(day_of_year from datetimes) as day_of_year,
        extract(week_of_year from datetimes) as week_of_year,
        extract(hour from datetimes) as hour,
        extract(minute from datetimes) as minute,
        extract(month from datetimes) as month,
        extract(second from datetimes) as second,
        extract(millisecond from datetimes) as millisecond,
        extract(microsecond from datetimes) as microsecond,
        extract(nanosecond from datetimes) as nanosecond,
        extract(quarter from datetimes) as quarter,
        extract(year from datetimes) as year,
        extract(unix_date from datetimes) as unix_date,
    FROM df
    """).collect()

    assert actual.to_pydict() == expected.to_pydict()


def test_date_comparison():
    date_df = daft.from_pydict({"date_str": ["2020-01-01", "2020-01-02", "2020-01-03"]})
    date_df = date_df.with_column("date", daft.col("date_str").str.to_date("%Y-%m-%d"))
    expected = date_df.filter(daft.col("date") == "2020-01-01").select("date").to_pydict()
    actual = daft.sql("select date from date_df where date == '2020-01-01'").to_pydict()
    assert actual == expected



from __future__ import annotations

import gcsfs
import pytest

from daft.daft import io_glob

BUCKET = "daft-public-data-gs"


def gcsfs_recursive_list(fs, path) -> list:
    all_results = []
    curr_level_result = fs.ls(path, detail=True)
    for item in curr_level_result:
        if item["type"] == "directory":
            new_path = f'gs://{item["name"]}'
            all_results.extend(gcsfs_recursive_list(fs, new_path))
            item["name"] += "/"
            all_results.append(item)
        else:
            all_results.append(item)
    return all_results


def compare_gcs_result(daft_ls_result: list, fsspec_result: list):
    daft_files = [(f["path"], f["type"].lower()) for f in daft_ls_result]
    gcsfs_files = [(f"gs://{f['name']}", f["type"]) for f in fsspec_result]

    # Remove all directories: our glob utilities don't return dirs
    gcsfs_files = [(path, type_) for path, type_ in gcsfs_files if type_ == "file"]

    # Remove size-0 directory marker files
    gcsfs_files = [(path, type_) for path, type_ in gcsfs_files if not path.endswith("/")]

    assert len(daft_files) == len(gcsfs_files)
    assert sorted(daft_files) == sorted(gcsfs_files)


@pytest.mark.integration()
@pytest.mark.parametrize(
    "path",
    [
        f"gs://{BUCKET}",
        f"gs://{BUCKET}/",
        f"gs://{BUCKET}/test_ls",
        f"gs://{BUCKET}/test_ls/",
        f"gs://{BUCKET}/test_ls//",
        f"gs://{BUCKET}/test_ls/paginated-1100-files/",
    ],
)
@pytest.mark.parametrize("recursive", [False, True])
@pytest.mark.parametrize("fanout_limit", [None, 1])
def test_gs_flat_directory_listing(gcs_public_config, path, recursive, fanout_limit):
    fs = gcsfs.GCSFileSystem()
    glob_path = path.rstrip("/") + "/**" if recursive else path
    daft_ls_result = io_glob(glob_path, io_config=gcs_public_config, fanout_limit=fanout_limit)
    fsspec_result = gcsfs_recursive_list(fs, path) if recursive else fs.ls(path, detail=True)
    compare_gcs_result(daft_ls_result, fsspec_result)


@pytest.mark.integration()
@pytest.mark.parametrize("recursive", [False, True])
def test_gs_single_file_listing(gcs_public_config, recursive):
    path = f"gs://{BUCKET}/test_ls/file.txt"
    fs = gcsfs.GCSFileSystem()
    daft_ls_result = io_glob(path, io_config=gcs_public_config)
    fsspec_result = gcsfs_recursive_list(fs, path) if recursive else fs.ls(path, detail=True)
    compare_gcs_result(daft_ls_result, fsspec_result)


@pytest.mark.integration()
def test_gs_notfound(gcs_public_config):
    path = f"gs://{BUCKET}/test_"
    with pytest.raises(FileNotFoundError, match=path):
        io_glob(path, io_config=gcs_public_config)



from __future__ import annotations

import pytest

from daft.expressions import col
from tests.conftest import assert_df_equals, get_tests_daft_runner_name


def skip_invalid_join_strategies(join_strategy):
    if get_tests_daft_runner_name() == "native":
        if join_strategy not in [None, "hash"]:
            pytest.skip("Native executor fails for these tests")


@pytest.mark.parametrize(
    "join_strategy",
    [None, "hash", "sort_merge", "sort_merge_aligned_boundaries", "broadcast"],
    indirect=True,
)
def test_simple_join(join_strategy, daft_df, service_requests_csv_pd_df, repartition_nparts, with_morsel_size):
    skip_invalid_join_strategies(join_strategy)
    daft_df = daft_df.repartition(repartition_nparts)
    daft_df_left = daft_df.select(col("Unique Key"), col("Borough"))
    daft_df_right = daft_df.select(col("Unique Key"), col("Created Date"))
    daft_df = daft_df_left.join(daft_df_right, col("Unique Key"), strategy=join_strategy)

    service_requests_csv_pd_df_left = service_requests_csv_pd_df[["Unique Key", "Borough"]]
    service_requests_csv_pd_df_right = service_requests_csv_pd_df[["Unique Key", "Created Date"]]
    service_requests_csv_pd_df = (
        service_requests_csv_pd_df_left.set_index("Unique Key")
        .join(service_requests_csv_pd_df_right.set_index("Unique Key"), how="inner")
        .reset_index()
    )
    daft_pd_df = daft_df.to_pandas()
    assert_df_equals(daft_pd_df, service_requests_csv_pd_df)


@pytest.mark.parametrize(
    "join_strategy",
    [None, "hash", "sort_merge", "sort_merge_aligned_boundaries", "broadcast"],
    indirect=True,
)
def test_simple_self_join(join_strategy, daft_df, service_requests_csv_pd_df, repartition_nparts, with_morsel_size):
    skip_invalid_join_strategies(join_strategy)
    daft_df = daft_df.repartition(repartition_nparts)
    daft_df = daft_df.select(col("Unique Key"), col("Borough"))

    daft_df = daft_df.join(daft_df, col("Unique Key"), strategy=join_strategy)

    service_requests_csv_pd_df = service_requests_csv_pd_df[["Unique Key", "Borough"]]
    service_requests_csv_pd_df = (
        service_requests_csv_pd_df.set_index("Unique Key")
        .join(
            service_requests_csv_pd_df.set_index("Unique Key"),
            how="inner",
            rsuffix="_right",
        )
        .reset_index()
    )
    service_requests_csv_pd_df = service_requests_csv_pd_df.rename({"Borough_right": "right.Borough"}, axis=1)
    daft_pd_df = daft_df.to_pandas()
    assert_df_equals(daft_pd_df, service_requests_csv_pd_df)


@pytest.mark.parametrize(
    "join_strategy",
    [None, "hash", "sort_merge", "sort_merge_aligned_boundaries", "broadcast"],
    indirect=True,
)
def test_simple_join_missing_rvalues(
    join_strategy, daft_df, service_requests_csv_pd_df, repartition_nparts, with_morsel_size
):
    skip_invalid_join_strategies(join_strategy)
    daft_df_right = daft_df.sort("Unique Key").limit(25).repartition(repartition_nparts)
    daft_df_left = daft_df.repartition(repartition_nparts)
    daft_df_left = daft_df_left.select(col("Unique Key"), col("Borough"))
    daft_df_right = daft_df_right.select(col("Unique Key"), col("Created Date")).sort(col("Unique Key"))
    daft_df = daft_df_left.join(daft_df_right, col("Unique Key"), strategy=join_strategy)

    service_requests_csv_pd_df_left = service_requests_csv_pd_df[["Unique Key", "Borough"]]
    service_requests_csv_pd_df_right = (
        service_requests_csv_pd_df[["Unique Key", "Created Date"]].sort_values(by="Unique Key").head(25)
    )
    service_requests_csv_pd_df = (
        service_requests_csv_pd_df_left.set_index("Unique Key")
        .join(service_requests_csv_pd_df_right.set_index("Unique Key"), how="inner")
        .reset_index()
    )
    daft_pd_df = daft_df.to_pandas()
    assert_df_equals(daft_pd_df, service_requests_csv_pd_df)


@pytest.mark.parametrize(
    "join_strategy",
    [None, "hash", "sort_merge", "sort_merge_aligned_boundaries", "broadcast"],
    indirect=True,
)
def test_simple_join_missing_lvalues(
    join_strategy, daft_df, service_requests_csv_pd_df, repartition_nparts, with_morsel_size
):
    skip_invalid_join_strategies(join_strategy)
    daft_df_right = daft_df.repartition(repartition_nparts)
    daft_df_left = daft_df.sort(col("Unique Key")).limit(25).repartition(repartition_nparts)
    daft_df_left = daft_df_left.select(col("Unique Key"), col("Borough"))
    daft_df_right = daft_df_right.select(col("Unique Key"), col("Created Date"))
    daft_df = daft_df_left.join(daft_df_right, col("Unique Key"), strategy=join_strategy)

    service_requests_csv_pd_df_left = (
        service_requests_csv_pd_df[["Unique Key", "Borough"]].sort_values(by="Unique Key").head(25)
    )
    service_requests_csv_pd_df_right = service_requests_csv_pd_df[["Unique Key", "Created Date"]]
    service_requests_csv_pd_df = (
        service_requests_csv_pd_df_left.set_index("Unique Key")
        .join(service_requests_csv_pd_df_right.set_index("Unique Key"), how="inner")
        .reset_index()
    )
    daft_pd_df = daft_df.to_pandas()
    assert_df_equals(daft_pd_df, service_requests_csv_pd_df)



from __future__ import annotations

import daft


def test_read_huggingface_datasets_doesnt_flae():
    from daft import DataType as dt

    # run it multiple times to ensure it doesn't fail
    for _ in range(10):
        df = daft.read_parquet("hf://datasets/huggingface/documentation-images")
        schema = df.schema()
        expected = daft.Schema.from_pydict({"image": dt.struct({"bytes": dt.binary(), "path": dt.string()})})
        assert schema == expected



from __future__ import annotations

import logging

from daft import refresh_logger


def setup_debug_logger(
    exclude_prefix: list[str] = [],
    daft_only: bool = True,
) -> None:
    logging.basicConfig(level="DEBUG")
    root_logger = logging.getLogger()
    root_logger.setLevel(logging.DEBUG)

    if daft_only:
        for handler in root_logger.handlers:
            handler.addFilter(lambda record: record.name.startswith("daft"))

    if exclude_prefix:
        for prefix in exclude_prefix:
            for handler in root_logger.handlers:
                handler.addFilter(lambda record: not record.name.startswith(prefix))

    refresh_logger()



from __future__ import annotations

import datetime
import math

import numpy as np
import pyarrow as pa
import pytest

from daft import DataType, col, from_pydict, utils
from daft.logical.schema import Schema
from daft.recordbatch import MicroPartition
from daft.series import Series
from tests.dataframe.test_aggregations import _assert_all_hashable
from tests.recordbatch import (
    daft_comparable_types,
    daft_floating_types,
    daft_nonnull_types,
    daft_null_types,
    daft_numeric_types,
    daft_string_types,
)

test_table_count_cases = [
    ([], {"count": [0]}),
    ([None], {"count": [0]}),
    ([None, None, None], {"count": [0]}),
    ([0], {"count": [1]}),
    ([None, 0, None, 0, None], {"count": [2]}),
]


@pytest.mark.parametrize(
    "mp",
    [
        MicroPartition.from_pydict(
            {"a": pa.array([], type=pa.int64()), "b": pa.array([], type=pa.string())}
        ),  # 1 empty table
        MicroPartition.empty(Schema.from_pyarrow_schema(pa.schema({"a": pa.int64(), "b": pa.string()}))),  # No tables
    ],
)
def test_multipartition_count_empty(mp):
    counted = mp.agg([col("a").count()])
    assert len(counted) == 1
    assert counted.to_pydict() == {"a": [0]}

    counted = mp.agg([col("a").count()], group_by=[col("b")])
    assert len(counted) == 0
    assert counted.to_pydict() == {"b": [], "a": []}


@pytest.mark.parametrize(
    "mp",
    [
        MicroPartition.from_pydict({"a": [1, None, 3, None], "b": ["a", "a", "b", "b"]}),  # 1 table
        MicroPartition.concat(
            [
                MicroPartition.from_pydict({"a": np.array([]).astype(np.int64), "b": pa.array([], type=pa.string())}),
                MicroPartition.from_pydict({"a": [1], "b": ["a"]}),
                MicroPartition.from_pydict({"a": [None, 3, None], "b": ["a", "b", "b"]}),
            ]
        ),  # 3 tables
    ],
)
def test_multipartition_count(mp):
    counted = mp.agg([col("a").count()])
    assert len(counted) == 1
    assert counted.to_pydict() == {"a": [2]}

    counted = mp.agg([col("a").count()], group_by=[col("b")])
    assert len(counted) == 2
    assert counted.to_pydict() == {"b": ["a", "b"], "a": [1, 1]}


@pytest.mark.parametrize("idx_dtype", daft_nonnull_types, ids=[f"{_}" for _ in daft_nonnull_types])
@pytest.mark.parametrize("case", test_table_count_cases, ids=[f"{_}" for _ in test_table_count_cases])
def test_table_count(idx_dtype, case) -> None:
    input, expected = case
    if idx_dtype == DataType.date():
        input = [datetime.date(2020 + x, 1 + x, 1 + x) if x is not None else None for x in input]
    daft_recordbatch = MicroPartition.from_pydict({"input": input})
    daft_recordbatch = daft_recordbatch.eval_expression_list([col("input").cast(idx_dtype)])
    daft_recordbatch = daft_recordbatch.eval_expression_list([col("input").alias("count").count()])

    res = daft_recordbatch.to_pydict()
    assert res == expected


@pytest.mark.parametrize("length", [0, 1, 10])
def test_table_count_nulltype(length) -> None:
    daft_recordbatch = MicroPartition.from_pydict({"input": [None] * length})
    daft_recordbatch = daft_recordbatch.eval_expression_list([col("input").cast(DataType.null())])
    daft_recordbatch = daft_recordbatch.eval_expression_list([col("input").alias("count").count()])

    res = daft_recordbatch.to_pydict()["count"]
    assert res == [0]


def test_table_count_pyobject() -> None:
    daft_recordbatch = MicroPartition.from_pydict({"objs": [object(), object(), None, object(), None]})
    daft_recordbatch = daft_recordbatch.eval_expression_list([col("objs").alias("count").count()])

    res = daft_recordbatch.to_pydict()["count"]
    assert res == [3]


test_table_minmax_numerics_cases = [
    ([], {"min": [None], "max": [None]}),
    ([None], {"min": [None], "max": [None]}),
    ([None, None, None], {"min": [None], "max": [None]}),
    ([5], {"min": [5], "max": [5]}),
    ([None, 5], {"min": [5], "max": [5]}),
    ([None, 1, 21, None, 21, 1, None], {"min": [1], "max": [21]}),
]


@pytest.mark.parametrize("idx_dtype", daft_numeric_types)
@pytest.mark.parametrize(
    "case", test_table_minmax_numerics_cases, ids=[f"{_}" for _ in test_table_minmax_numerics_cases]
)
def test_table_minmax_numerics(idx_dtype, case) -> None:
    input, expected = case
    daft_recordbatch = MicroPartition.from_pydict({"input": input})
    daft_recordbatch = daft_recordbatch.eval_expression_list([col("input").cast(idx_dtype)])
    daft_recordbatch = daft_recordbatch.eval_expression_list(
        [
            col("input").alias("min").min(),
            col("input").alias("max").max(),
        ]
    )

    res = daft_recordbatch.to_pydict()
    assert res == expected


test_table_minmax_string_cases = [
    ([], {"min": [None], "max": [None]}),
    ([None], {"min": [None], "max": [None]}),
    ([None, None, None], {"min": [None], "max": [None]}),
    (["abc"], {"min": ["abc"], "max": ["abc"]}),
    ([None, "abc"], {"min": ["abc"], "max": ["abc"]}),
    ([None, "aaa", None, "b", "c", "aaa", None], {"min": ["aaa"], "max": ["c"]}),
]


@pytest.mark.parametrize("idx_dtype", daft_string_types)
@pytest.mark.parametrize("case", test_table_minmax_string_cases, ids=[f"{_}" for _ in test_table_minmax_string_cases])
def test_table_minmax_string(idx_dtype, case) -> None:
    input, expected = case
    daft_recordbatch = MicroPartition.from_pydict({"input": input})
    daft_recordbatch = daft_recordbatch.eval_expression_list([col("input").cast(idx_dtype)])
    daft_recordbatch = daft_recordbatch.eval_expression_list(
        [
            col("input").alias("min").min(),
            col("input").alias("max").max(),
        ]
    )

    res = daft_recordbatch.to_pydict()
    assert res == expected


test_table_minmax_bool_cases = [
    ([], {"min": [None], "max": [None]}),
    ([None], {"min": [None], "max": [None]}),
    ([None, None, None], {"min": [None], "max": [None]}),
    ([False, True], {"min": [False], "max": [True]}),
    ([None, None, True, None], {"min": [True], "max": [True]}),
]


@pytest.mark.parametrize("case", test_table_minmax_bool_cases, ids=[f"{_}" for _ in test_table_minmax_bool_cases])
def test_table_minmax_bool(case) -> None:
    input, expected = case
    daft_recordbatch = MicroPartition.from_pydict({"input": input})
    daft_recordbatch = daft_recordbatch.eval_expression_list([col("input").cast(DataType.bool())])
    daft_recordbatch = daft_recordbatch.eval_expression_list(
        [
            col("input").alias("min").min(),
            col("input").alias("max").max(),
        ]
    )

    res = daft_recordbatch.to_pydict()
    assert res == expected


test_table_minmax_binary_cases = [
    ([], {"min": [None], "max": [None]}),
    ([None], {"min": [None], "max": [None]}),
    ([None, None, None], {"min": [None], "max": [None]}),
    ([b"1"], {"min": [b"1"], "max": [b"1"]}),
    ([None, b"1"], {"min": [b"1"], "max": [b"1"]}),
    ([b"a", b"b", b"c", b"a"], {"min": [b"a"], "max": [b"c"]}),
]


@pytest.mark.parametrize("case", test_table_minmax_binary_cases, ids=[f"{_}" for _ in test_table_minmax_binary_cases])
@pytest.mark.parametrize("type", [pa.binary(), pa.binary(1)])
def test_table_minmax_binary(case, type) -> None:
    input, expected = case
    daft_recordbatch = MicroPartition.from_arrow(pa.table({"input": pa.array(input, type=type)}))
    daft_recordbatch = daft_recordbatch.eval_expression_list(
        [
            col("input").alias("min").min(),
            col("input").alias("max").max(),
        ]
    )

    res = daft_recordbatch.to_pydict()
    assert res == expected


test_table_sum_mean_cases = [
    ([], {"sum": [None], "mean": [None]}),
    ([None], {"sum": [None], "mean": [None]}),
    ([None, None, None], {"sum": [None], "mean": [None]}),
    ([0], {"sum": [0], "mean": [0]}),
    ([1], {"sum": [1], "mean": [1]}),
    ([None, 3, None, None, 1, 2, 0, None], {"sum": [6], "mean": [1.5]}),
]


@pytest.mark.parametrize("idx_dtype", daft_numeric_types, ids=[f"{_}" for _ in daft_numeric_types])
@pytest.mark.parametrize("case", test_table_sum_mean_cases, ids=[f"{_}" for _ in test_table_sum_mean_cases])
def test_table_sum_mean(idx_dtype, case) -> None:
    input, expected = case
    daft_recordbatch = MicroPartition.from_pydict({"input": input})
    daft_recordbatch = daft_recordbatch.eval_expression_list([col("input").cast(idx_dtype)])
    daft_recordbatch = daft_recordbatch.eval_expression_list(
        [
            col("input").alias("sum").sum(),
            col("input").alias("mean").mean(),
        ]
    )

    res = daft_recordbatch.to_pydict()
    assert res == expected


@pytest.mark.parametrize("nptype", [np.uint8, np.uint16, np.uint32, np.int8, np.int16, np.int32])
def test_table_sum_upcast(nptype) -> None:
    """Tests correctness, including type upcasting, of sum aggregations."""
    daft_recordbatch = MicroPartition.from_pydict(
        {
            "maxes": np.full(128, fill_value=np.iinfo(nptype).max, dtype=nptype),
            "mins": np.full(128, fill_value=np.iinfo(nptype).min, dtype=nptype),
        }
    )
    daft_recordbatch = daft_recordbatch.eval_expression_list([col("maxes").sum(), col("mins").sum()])
    pydict = daft_recordbatch.to_pydict()
    assert pydict["maxes"] == [128 * np.iinfo(nptype).max]
    assert pydict["mins"] == [128 * np.iinfo(nptype).min]


def test_table_sum_badtype() -> None:
    daft_recordbatch = MicroPartition.from_pydict({"a": ["str1", "str2"]})
    with pytest.raises(ValueError):
        daft_recordbatch = daft_recordbatch.eval_expression_list([col("a").sum()])


test_micropartition_any_value_cases = [
    (
        MicroPartition.from_pydict({"a": [None, 1, None, None], "b": ["a", "a", "b", "b"]}),  # 1 table
        {"a": True, "b": True},
        {"a": False, "b": True},
    ),
    (
        MicroPartition.concat(
            [
                MicroPartition.from_pydict({"a": np.array([]).astype(np.int64), "b": pa.array([], type=pa.string())}),
                MicroPartition.from_pydict({"a": [None, 3, None], "b": ["a", "b", "b"]}),
            ]
        ),  # 2 tables
        {"a": True, "b": True},
        {"a": True, "b": False},
    ),
    (
        MicroPartition.concat(
            [
                MicroPartition.from_pydict({"a": np.array([]).astype(np.int64), "b": pa.array([], type=pa.string())}),
                MicroPartition.from_pydict({"a": [None, 3, None], "b": ["a", "b", "b"]}),
                MicroPartition.from_pydict({"a": [1], "b": ["a"]}),
            ]
        ),  # 3 tables
        {"a": True, "b": True},
        {"a": False, "b": False},
    ),
]


@pytest.mark.parametrize("mp,expected_nulls,expected_no_nulls", test_micropartition_any_value_cases)
def test_micropartition_any_value(mp, expected_nulls, expected_no_nulls):
    any_values = mp.agg([col("a").any_value(False)], group_by=[col("b")]).to_pydict()
    assert len(any_values["b"]) == len(expected_nulls)
    for k, v in zip(any_values["b"], any_values["a"]):
        assert expected_nulls[k] or v is not None

    any_values = mp.agg([col("a").any_value(True)], group_by=[col("b")]).to_pydict()
    assert len(any_values["b"]) == len(expected_no_nulls)
    for k, v in zip(any_values["b"], any_values["a"]):
        assert expected_no_nulls[k] or v is not None


test_table_any_value_cases = [
    ({"a": [1], "b": ["a"]}, {"a": False}, {"a": False}),
    ({"a": [None], "b": ["a"]}, {"a": True}, {"a": True}),
    ({"a": [None, 1], "b": ["a", "a"]}, {"a": True}, {"a": False}),
    ({"a": [1, None, 2], "b": ["a", "b", "b"]}, {"a": False, "b": True}, {"a": False, "b": False}),
]


@pytest.mark.parametrize("case,expected_nulls,expected_no_nulls", test_table_any_value_cases)
def test_table_any_value(case, expected_nulls, expected_no_nulls):
    daft_recordbatch = MicroPartition.from_pydict(case)

    any_values = daft_recordbatch.agg([col("a").any_value(False)], group_by=[col("b")]).to_pydict()
    assert len(any_values["b"]) == len(expected_nulls)
    for k, v in zip(any_values["b"], any_values["a"]):
        assert expected_nulls[k] or v is not None

    any_values = daft_recordbatch.agg([col("a").any_value(True)], group_by=[col("b")]).to_pydict()
    assert len(any_values["b"]) == len(expected_no_nulls)
    for k, v in zip(any_values["b"], any_values["a"]):
        assert expected_no_nulls[k] or v is not None


test_table_agg_global_cases = [
    (
        [],
        {
            "count": [0],
            "sum": [None],
            "mean": [None],
            "min": [None],
            "max": [None],
            "list": [[]],
            "set": [[]],
        },
    ),
    (
        [None],
        {
            "count": [0],
            "sum": [None],
            "mean": [None],
            "min": [None],
            "max": [None],
            "list": [[None]],
            "set": [[]],
        },
    ),
    (
        [None, None, None],
        {
            "count": [0],
            "sum": [None],
            "mean": [None],
            "min": [None],
            "max": [None],
            "list": [[None, None, None]],
            "set": [[]],
        },
    ),
    (
        [None, 3, None, None, 1, 2, 0, None],
        {
            "count": [4],
            "sum": [6],
            "mean": [1.5],
            "min": [0],
            "max": [3],
            "list": [[None, 3, None, None, 1, 2, 0, None]],
            "set": [[0, 1, 2, 3]],
        },
    ),
]


@pytest.mark.parametrize("case", test_table_agg_global_cases, ids=[f"{_}" for _ in test_table_agg_global_cases])
def test_table_agg_global(case) -> None:
    """Test that global aggregation works at the API layer."""
    input, expected = case
    daft_recordbatch = MicroPartition.from_pydict({"input": input})
    daft_recordbatch = daft_recordbatch.agg(
        [
            col("input").cast(DataType.int32()).alias("count").count(),
            col("input").cast(DataType.int32()).alias("sum").sum(),
            col("input").cast(DataType.int32()).alias("mean").mean(),
            col("input").cast(DataType.int32()).alias("min").min(),
            col("input").cast(DataType.int32()).alias("max").max(),
            col("input").cast(DataType.int32()).alias("list").agg_list(),
            col("input").cast(DataType.int32()).alias("set").agg_set(),
        ]
    )

    result = daft_recordbatch.to_pydict()

    # Handle list and set results separately
    res_list = result.pop("list")
    exp_list = expected.pop("list")
    res_set = result.pop("set")
    exp_set = expected.pop("set")

    # Check regular aggregations
    for key, value in expected.items():
        assert result[key] == value

    # Check list result
    assert len(res_list) == 1
    assert res_list[0] == exp_list[0]

    # Check set without nulls
    assert len(res_set) == 1
    _assert_all_hashable(res_set[0], "test_table_agg_global")
    assert len(res_set[0]) == len(set(x for x in res_set[0] if x is not None)), "Result should contain no duplicates"
    assert set(x for x in res_set[0] if x is not None) == set(
        x for x in exp_set[0] if x is not None
    ), "Sets should contain same non-null elements"
    assert None not in res_set[0], "Result should not contain nulls"


@pytest.mark.parametrize(
    "groups_and_aggs",
    [
        (["col_A"], ["col_B"]),
        (["col_A", "col_B"], []),
    ],
)
def test_table_agg_groupby_empty(groups_and_aggs) -> None:
    groups, aggs = groups_and_aggs
    daft_recordbatch = MicroPartition.from_pydict({"col_A": [], "col_B": []})
    daft_recordbatch = daft_recordbatch.agg(
        [col(a).count() for a in aggs],
        [col(g).cast(DataType.int32()) for g in groups],
    )
    res = daft_recordbatch.to_pydict()

    assert res == {"col_A": [], "col_B": []}


test_table_agg_groupby_cases = [
    {
        # Group by strings.
        "groups": ["name"],
        "aggs": [
            col("cookies").alias("sum").sum(),
            col("name").alias("count").count(),
            col("cookies").alias("list").agg_list(),
            col("cookies").alias("set").agg_set(),
        ],
        "expected": {
            "name": ["Alice", "Bob", None],
            "sum": [None, 10, 7],
            "count": [4, 4, 0],
            "list": [[None] * 4, [None, None, 5, 5], [None, 5, None, 2]],
            "set": [set(), {5}, {2, 5}],
        },
    },
    {
        # Group by numbers.
        "groups": ["cookies"],
        "aggs": [col("name").alias("count").count()],
        "expected": {"cookies": [2, 5, None], "count": [0, 2, 6]},
    },
    {
        # Group by multicol.
        "groups": ["name", "cookies"],
        "aggs": [col("name").alias("count").count()],
        "expected": {
            "name": ["Alice", "Bob", "Bob", None, None, None],
            "cookies": [None, 5, None, 2, 5, None],
            "count": [4, 2, 2, 0, 0, 0],
        },
    },
]


@pytest.mark.parametrize(
    "case", test_table_agg_groupby_cases, ids=[f"{case['groups']}" for case in test_table_agg_groupby_cases]
)
def test_table_agg_groupby(case) -> None:
    values = [
        ("Bob", None),
        ("Bob", None),
        ("Bob", 5),
        ("Bob", 5),
        (None, None),
        (None, 5),
        (None, None),
        (None, 2),
        ("Alice", None),
        ("Alice", None),
        ("Alice", None),
        ("Alice", None),
    ]
    daft_recordbatch = MicroPartition.from_pydict(
        {
            "name": [_[0] for _ in values],
            "cookies": [_[1] for _ in values],
        }
    )
    # Sort by grouping columns after aggregation
    daft_recordbatch = daft_recordbatch.agg(
        [aggexpr for aggexpr in case["aggs"]],
        [col(group) for group in case["groups"]],
    )
    result = daft_recordbatch.sort([col(group) for group in case["groups"]]).to_pydict()
    expected = case["expected"]

    # Compare non-set columns normally
    for key in result:
        if key == "set":
            # Compare set columns by converting to sets
            assert len(result[key]) == len(expected[key]), f"Length mismatch in column {key}"
            for res, exp in zip(result[key], expected[key]):
                _assert_all_hashable(res, "test_table_agg_groupby")
                assert set(res) == exp, f"Set mismatch in column {key}"
        else:
            assert result[key] == expected[key], f"Mismatch in column {key}"


@pytest.mark.parametrize("dtype", daft_comparable_types, ids=[f"{_}" for _ in daft_comparable_types])
def test_groupby_all_nulls(dtype) -> None:
    daft_recordbatch = MicroPartition.from_pydict(
        {
            "group": Series.from_pylist([None, None, None]).cast(dtype),
            "cookies": [1, 2, 3],
        }
    )
    result_table = daft_recordbatch.agg([col("cookies").sum()], group_by=[col("group")])
    assert result_table.to_pydict() == {"group": [None], "cookies": [6]}


@pytest.mark.parametrize(
    "dtype",
    daft_numeric_types + daft_string_types + [DataType.bool()],
    ids=[f"{_}" for _ in daft_numeric_types + daft_string_types + [DataType.bool()]],
)
def test_groupby_numeric_string_bool_some_nulls(dtype) -> None:
    daft_recordbatch = MicroPartition.from_pydict(
        {
            "group": Series.from_pylist([1, 1, None]).cast(dtype),
            "cookies": [2, 2, 3],
        }
    )
    result_table = daft_recordbatch.agg([col("cookies").sum()], group_by=[col("group")])
    expected_table = MicroPartition.from_pydict(
        {
            "group": Series.from_pylist([1, None]).cast(dtype),
            "cookies": [4, 3],
        }
    )

    assert set(utils.freeze(utils.pydict_to_rows(result_table.to_pydict()))) == set(
        utils.freeze(utils.pydict_to_rows(expected_table.to_pydict()))
    )


@pytest.mark.parametrize(
    "dtype",
    daft_numeric_types + daft_string_types + [DataType.bool()],
    ids=[f"{_}" for _ in daft_numeric_types + daft_string_types + [DataType.bool()]],
)
def test_groupby_numeric_string_bool_no_nulls(dtype) -> None:
    daft_recordbatch = MicroPartition.from_pydict(
        {
            "group": Series.from_pylist([1, 0, 1, 0]).cast(dtype),
            "cookies": [1, 2, 2, 3],
        }
    )
    result_table = daft_recordbatch.agg([col("cookies").sum()], group_by=[col("group")])
    expected_table = MicroPartition.from_pydict(
        {
            "group": Series.from_pylist([0, 1]).cast(dtype),
            "cookies": [5, 3],
        }
    )

    assert set(utils.freeze(utils.pydict_to_rows(result_table.to_pydict()))) == set(
        utils.freeze(utils.pydict_to_rows(expected_table.to_pydict()))
    )


@pytest.mark.parametrize("type", [pa.binary(), pa.binary(1)])
@pytest.mark.parametrize(
    "agg, expected",
    [
        (col("cookies").max(), [b"2", b"4"]),
        (col("cookies").min(), [b"1", b"3"]),
    ],
)
def test_groupby_binary_bool_some_nulls(type, agg, expected) -> None:
    daft_recordbatch = MicroPartition.from_pydict(
        {
            "group": Series.from_arrow(pa.array([b"1", b"1", None, None], type=type)),
            "cookies": Series.from_arrow(pa.array([b"1", b"2", b"3", b"4"], type=type)),
        }
    )
    result_table = daft_recordbatch.agg([agg], group_by=[col("group")])
    expected_table = MicroPartition.from_pydict(
        {
            "group": Series.from_pylist([b"1", None]),
            "cookies": expected,
        }
    )

    assert set(utils.freeze(utils.pydict_to_rows(result_table.to_pydict()))) == set(
        utils.freeze(utils.pydict_to_rows(expected_table.to_pydict()))
    )


@pytest.mark.parametrize("type", [pa.binary(), pa.binary(1)])
@pytest.mark.parametrize(
    "agg, expected",
    [
        (col("cookies").max(), [b"4", b"3"]),
        (col("cookies").min(), [b"2", b"1"]),
    ],
)
def test_groupby_binary_no_nulls(type, agg, expected) -> None:
    daft_recordbatch = MicroPartition.from_pydict(
        {
            "group": Series.from_arrow(pa.array([b"1", b"0", b"1", b"0"], type=type)),
            "cookies": Series.from_arrow(pa.array([b"1", b"2", b"3", b"4"], type=type)),
        }
    )
    result_table = daft_recordbatch.agg([agg], group_by=[col("group")])
    expected_table = MicroPartition.from_pydict(
        {
            "group": Series.from_pylist([b"0", b"1"]),
            "cookies": expected,
        }
    )

    assert set(utils.freeze(utils.pydict_to_rows(result_table.to_pydict()))) == set(
        utils.freeze(utils.pydict_to_rows(expected_table.to_pydict()))
    )


@pytest.mark.parametrize(
    "dtype",
    daft_floating_types,
    ids=[f"{_}" for _ in daft_floating_types],
)
def test_groupby_floats_nan(dtype) -> None:
    NAN = float("nan")
    INF = float("inf")

    daft_recordbatch = MicroPartition.from_pydict(
        {
            "group": Series.from_pylist([None, 1.0, NAN, 5 * NAN, -1 * NAN, -NAN, 1.0, None, INF, -INF, INF]).cast(
                dtype
            ),
            "cookies": [0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10],
        }
    )
    result_table = daft_recordbatch.agg([col("cookies").count()], group_by=[col("group")])
    expected_table = MicroPartition.from_pydict(
        {
            "group": Series.from_pylist([None, 1.0, NAN, -INF, INF]).cast(dtype),
            "cookies": [2, 2, 4, 1, 2],
        }
    )
    # have to sort and compare since `utils.pydict_to_rows` doesn't work on NaNs
    for result_col, expected_col in zip(
        result_table.sort([col("group")]).to_pydict(), expected_table.sort([col("group")]).to_pydict()
    ):
        for r, e in zip(result_col, expected_col):
            assert (r == e) or (math.isnan(r) and math.isnan(e))


def test_groupby_timestamp() -> None:
    daft_recordbatch = MicroPartition.from_pydict(
        {
            "group": Series.from_pylist(
                [
                    datetime.datetime(2020, 1, 1, 0, 0, 0),
                    datetime.datetime(2020, 1, 1, 0, 30, 0),
                    datetime.datetime(2020, 1, 1, 0, 59, 59),
                    datetime.datetime(2020, 1, 1, 1, 0, 0),
                    datetime.datetime(2020, 1, 1, 1, 30, 0),
                    datetime.datetime(2020, 1, 1, 1, 59, 59),
                    datetime.datetime(2020, 1, 1, 2, 0, 0),
                    datetime.datetime(2020, 1, 1, 2, 30, 0),
                    datetime.datetime(2020, 1, 1, 2, 59, 59),
                ]
            ),
            "value": [1, 1, 1, 2, 2, 2, 3, 3, 3],
        }
    )
    result_table = daft_recordbatch.agg([col("value").sum()], group_by=[col("group").dt.truncate("1 hour")])
    expected_table = MicroPartition.from_pydict(
        {
            "group": Series.from_pylist(
                [
                    datetime.datetime(2020, 1, 1, 0, 0, 0),
                    datetime.datetime(2020, 1, 1, 1, 0, 0),
                    datetime.datetime(2020, 1, 1, 2, 0, 0),
                ]
            ),
            "value": [3, 6, 9],
        }
    )

    assert set(utils.freeze(utils.pydict_to_rows(result_table.to_pydict()))) == set(
        utils.freeze(utils.pydict_to_rows(expected_table.to_pydict()))
    )


@pytest.mark.parametrize(
    "dtype", daft_nonnull_types + daft_null_types, ids=[f"{_}" for _ in daft_nonnull_types + daft_null_types]
)
def test_global_list_aggs(dtype) -> None:
    input = [None, 0, 1, 2, None, 4]
    if dtype == DataType.date():
        input = [datetime.date(2020 + x, 1 + x, 1 + x) if x is not None else None for x in input]
    daft_recordbatch = MicroPartition.from_pydict({"input": input})
    daft_recordbatch = daft_recordbatch.eval_expression_list([col("input").cast(dtype)])
    result = daft_recordbatch.eval_expression_list([col("input").alias("list").agg_list()])
    assert result.get_column_by_name("list").datatype() == DataType.list(dtype)
    assert result.to_pydict() == {"list": [daft_recordbatch.to_pydict()["input"]]}


def test_global_pyobj_list_aggs() -> None:
    input = [object(), object(), object()]
    table = MicroPartition.from_pydict({"input": input})
    result = table.eval_expression_list([col("input").alias("list").agg_list()])
    assert result.get_column_by_name("list").datatype() == DataType.python()
    assert result.to_pydict()["list"][0] == input


def test_global_list_list_aggs() -> None:
    input = [[1], [2, 3, 4], [5, None], [], None]
    table = MicroPartition.from_pydict({"input": input})
    result = table.eval_expression_list([col("input").alias("list").agg_list()])
    assert result.get_column_by_name("list").datatype() == DataType.list(DataType.list(DataType.int64()))
    assert result.to_pydict()["list"][0] == input


def test_global_fixed_size_list_list_aggs() -> None:
    input = Series.from_pylist([[1, 2], [3, 4], [5, None], None]).cast(DataType.fixed_size_list(DataType.int64(), 2))
    table = MicroPartition.from_pydict({"input": input})
    result = table.eval_expression_list([col("input").alias("list").agg_list()])
    assert result.get_column_by_name("list").datatype() == DataType.list(DataType.fixed_size_list(DataType.int64(), 2))
    assert result.to_pydict()["list"][0] == [[1, 2], [3, 4], [5, None], None]


def test_global_struct_list_aggs() -> None:
    input = [{"a": 1, "b": 2}, {"a": 3, "b": None}, None]
    table = MicroPartition.from_pydict({"input": input})
    result = table.eval_expression_list([col("input").alias("list").agg_list()])
    assert result.get_column_by_name("list").datatype() == DataType.list(
        DataType.struct({"a": DataType.int64(), "b": DataType.int64()})
    )
    assert result.to_pydict()["list"][0] == input


@pytest.mark.parametrize(
    "dtype", daft_nonnull_types + daft_null_types, ids=[f"{_}" for _ in daft_nonnull_types + daft_null_types]
)
def test_grouped_list_aggs(dtype) -> None:
    groups = [None, 1, None, 1, 2, 2]
    input = [None, 0, 1, 2, None, 4]
    expected_idx = [[1, 3], [4, 5], [0, 2]]

    if dtype == DataType.date():
        input = [datetime.date(2020 + x, 1 + x, 1 + x) if x is not None else None for x in input]
    daft_recordbatch = MicroPartition.from_pydict({"groups": groups, "input": input})
    daft_recordbatch = daft_recordbatch.eval_expression_list([col("groups"), col("input").cast(dtype)])
    result = daft_recordbatch.agg([col("input").alias("list").agg_list()], group_by=[col("groups")]).sort(
        [col("groups")]
    )
    assert result.get_column_by_name("list").datatype() == DataType.list(dtype)

    input_as_dtype = daft_recordbatch.get_column_by_name("input").to_pylist()
    expected_groups = [[input_as_dtype[i] for i in group] for group in expected_idx]

    assert result.to_pydict() == {"groups": [1, 2, None], "list": expected_groups}


def test_grouped_pyobj_list_aggs() -> None:
    groups = [None, 1, None, 1, 2, 2]
    input = [None, object(), object(), object(), None, object()]
    expected_idx = [[1, 3], [4, 5], [0, 2]]

    daft_recordbatch = MicroPartition.from_pydict({"groups": groups, "input": input})
    result = daft_recordbatch.agg([col("input").alias("list").agg_list()], group_by=[col("groups")]).sort(
        [col("groups")]
    )
    expected_groups = [[input[i] for i in group] for group in expected_idx]
    assert result.to_pydict() == {"groups": [1, 2, None], "list": expected_groups}


def test_grouped_list_list_aggs() -> None:
    groups = [None, 1, None, 1, 2, 2]
    input = [[1], [2, 3, 4], [5, None], None, [], [8, 9]]
    expected_idx = [[1, 3], [4, 5], [0, 2]]

    daft_recordbatch = MicroPartition.from_pydict({"groups": groups, "input": input})
    daft_recordbatch = daft_recordbatch.eval_expression_list([col("groups"), col("input")])
    result = daft_recordbatch.agg([col("input").alias("list").agg_list()], group_by=[col("groups")]).sort(
        [col("groups")]
    )
    assert result.get_column_by_name("list").datatype() == DataType.list(DataType.list(DataType.int64()))

    input_as_dtype = daft_recordbatch.get_column_by_name("input").to_pylist()
    expected_groups = [[input_as_dtype[i] for i in group] for group in expected_idx]

    assert result.to_pydict() == {"groups": [1, 2, None], "list": expected_groups}


def test_grouped_fixed_size_list_list_aggs() -> None:
    groups = [None, 1, None, 1, 2, 2]
    input = Series.from_pylist([[1, 2], [3, 4], [5, None], None, [6, 7], [8, 9]]).cast(
        DataType.fixed_size_list(DataType.int64(), 2)
    )
    expected_idx = [[1, 3], [4, 5], [0, 2]]

    daft_recordbatch = MicroPartition.from_pydict({"groups": groups, "input": input})
    daft_recordbatch = daft_recordbatch.eval_expression_list([col("groups"), col("input")])
    result = daft_recordbatch.agg([col("input").alias("list").agg_list()], group_by=[col("groups")]).sort(
        [col("groups")]
    )
    assert result.get_column_by_name("list").datatype() == DataType.list(DataType.fixed_size_list(DataType.int64(), 2))

    input_as_dtype = daft_recordbatch.get_column_by_name("input").to_pylist()
    expected_groups = [[input_as_dtype[i] for i in group] for group in expected_idx]

    assert result.to_pydict() == {"groups": [1, 2, None], "list": expected_groups}


def test_grouped_struct_list_aggs() -> None:
    groups = [None, 1, None, 1, 2, 2]
    input = [{"x": 1, "y": 2}, {"x": 3, "y": 4}, {"x": 5, "y": None}, None, {"x": 6, "y": 7}, {"x": 8, "y": 9}]
    expected_idx = [[1, 3], [4, 5], [0, 2]]

    daft_recordbatch = MicroPartition.from_pydict({"groups": groups, "input": input})
    daft_recordbatch = daft_recordbatch.eval_expression_list([col("groups"), col("input")])
    result = daft_recordbatch.agg([col("input").alias("list").agg_list()], group_by=[col("groups")]).sort(
        [col("groups")]
    )
    assert result.get_column_by_name("list").datatype() == DataType.list(
        DataType.struct({"x": DataType.int64(), "y": DataType.int64()})
    )

    input_as_dtype = daft_recordbatch.get_column_by_name("input").to_pylist()
    expected_groups = [[input_as_dtype[i] for i in group] for group in expected_idx]

    assert result.to_pydict() == {"groups": [1, 2, None], "list": expected_groups}


def test_list_aggs_empty() -> None:
    daft_recordbatch = MicroPartition.from_pydict({"col_A": [], "col_B": []})
    daft_recordbatch = daft_recordbatch.agg(
        [col("col_A").cast(DataType.int32()).alias("list").agg_list()],
        group_by=[col("col_B")],
    )
    assert daft_recordbatch.get_column_by_name("list").datatype() == DataType.list(DataType.int32())
    res = daft_recordbatch.to_pydict()

    assert res == {"col_B": [], "list": []}


@pytest.mark.parametrize(
    "dtype", daft_nonnull_types + daft_null_types, ids=[f"{_}" for _ in daft_nonnull_types + daft_null_types]
)
@pytest.mark.parametrize("with_null", [False, True])
def test_global_concat_aggs(dtype, with_null) -> None:
    input = [None, 0, 1, 2, None, 4]

    if dtype == DataType.date():
        input = [datetime.date(2020 + x, 1 + x, 1 + x) if x is not None else None for x in input]

    input = [[x] for x in input]
    if with_null:
        input += [None]

    daft_recordbatch = MicroPartition.from_pydict({"input": input}).eval_expression_list(
        [col("input").cast(DataType.list(dtype))]
    )
    concated = daft_recordbatch.agg([col("input").alias("concat").agg_concat()])
    assert concated.get_column_by_name("concat").datatype() == DataType.list(dtype)

    input_as_dtype = daft_recordbatch.get_column_by_name("input").to_pylist()
    # We should ignore Null Array elements when performing the concat agg
    expected = [[val[0] for val in input_as_dtype if val is not None]]
    assert concated.to_pydict() == {"concat": expected}


def test_global_concat_aggs_pyobj() -> None:
    expected = [object(), object(), None, None, object()]
    input = [
        [expected[0], expected[1]],
        None,
        [expected[2]],
        [expected[3], expected[4]],
    ]

    table = MicroPartition.from_pydict({"input": input})
    concatted = table.agg([col("input").alias("concat").agg_concat()])
    assert concatted.get_column_by_name("concat").datatype() == DataType.python()
    assert concatted.to_pydict()["concat"] == [expected]


@pytest.mark.parametrize(
    "dtype", daft_nonnull_types + daft_null_types, ids=[f"{_}" for _ in daft_nonnull_types + daft_null_types]
)
def test_grouped_concat_aggs(dtype) -> None:
    input = [None, 0, 1, 2, None, 4]

    if dtype == DataType.date():
        input = [datetime.date(2020 + x, 1 + x, 1 + x) if x is not None else None for x in input]

    input = [[x] for x in input] + [None]
    groups = [1, 2, 3, 4, 5, 6, 7]
    daft_recordbatch = MicroPartition.from_pydict({"groups": groups, "input": input}).eval_expression_list(
        [col("groups"), col("input").cast(DataType.list(dtype))]
    )
    concat_grouped = daft_recordbatch.agg(
        [col("input").alias("concat").agg_concat()], group_by=[col("groups") % 2]
    ).sort([col("groups")])
    assert concat_grouped.get_column_by_name("concat").datatype() == DataType.list(dtype)

    input_as_dtype = daft_recordbatch.get_column_by_name("input").to_pylist()
    # We should ignore Null Array elements when performing the concat agg
    expected_groups = [
        [input_as_dtype[i][0] for i in group if input_as_dtype[i] is not None] for group in [[1, 3, 5], [0, 2, 4, 6]]
    ]
    assert concat_grouped.to_pydict() == {"groups": [0, 1], "concat": expected_groups}


def test_grouped_concat_aggs_pyobj() -> None:
    objects = [object(), object(), object(), object()]
    input = [
        [objects[0], objects[1]],
        None,
        None,
        [objects[2]],
        [None, objects[3]],
    ]

    table = MicroPartition.from_pydict({"input": input, "groups": [1, 2, 3, 3, 4]})
    concatted = table.agg([col("input").alias("concat").agg_concat()], group_by=[col("groups")]).sort([col("groups")])
    assert concatted.get_column_by_name("concat").datatype() == DataType.python()
    assert concatted.to_pydict() == {
        "groups": [1, 2, 3, 4],
        "concat": [
            [objects[0], objects[1]],
            [],
            [objects[2]],
            [None, objects[3]],
        ],
    }


def test_concat_aggs_empty() -> None:
    daft_recordbatch = MicroPartition.from_pydict({"col_A": [], "col_B": []})
    daft_recordbatch = daft_recordbatch.agg(
        [col("col_A").cast(DataType.list(DataType.int32())).alias("concat").agg_concat()],
        group_by=[col("col_B")],
    )

    assert daft_recordbatch.get_column_by_name("concat").datatype() == DataType.list(DataType.int32())
    res = daft_recordbatch.to_pydict()

    assert res == {"col_B": [], "concat": []}


@pytest.mark.parametrize("dtype", daft_numeric_types)
def test_groupby_list(dtype) -> None:
    df = from_pydict(
        {
            "a": [[1, 2, 3], [1, 2, 3], [1, 2], [], [1, 2, 3], [], [1, 2]],
            "b": [0, 1, 2, 3, 4, 5, 6],
        }
    ).with_column("a", col("a").cast(DataType.list(dtype)))
    res = df.groupby("a").agg_list("b").to_pydict()
    expected = [[0, 1, 4], [2, 6], [3, 5]]
    for lt in expected:
        assert lt in res["b"]


@pytest.mark.parametrize("dtype", daft_numeric_types)
def test_groupby_fixed_size_list(dtype) -> None:
    df = from_pydict(
        {
            "a": [
                [1, 2, 3],
                [1, 2, 3],
                [1, 2, 4],
                [3, 2, 1],
                [1, 2, 3],
                [3, 2, 1],
                [1, 2, 4],
            ],
            "b": [0, 1, 2, 3, 4, 5, 6],
        }
    ).with_column("a", col("a").cast(DataType.fixed_size_list(dtype, 3)))
    res = df.groupby("a").agg_list("b").to_pydict()
    expected = [[0, 1, 4], [2, 6], [3, 5]]
    for lt in expected:
        assert lt in res["b"]


@pytest.mark.parametrize("dtype", daft_numeric_types)
def test_groupby_struct(dtype) -> None:
    df = from_pydict(
        {
            "a": [
                {"c": 1, "d": "hi"},
                {"c": 1, "d": "hi"},
                {"c": 1, "d": "hello"},
                {"c": 2, "d": "hello"},
                {"c": 1, "d": "hi"},
                {"c": 2, "d": "hello"},
                {"c": 1, "d": "hello"},
            ],
            "b": [0, 1, 2, 3, 4, 5, 6],
        }
    ).with_column("a", col("a").cast(DataType.struct({"c": dtype, "d": DataType.string()})))
    res = df.groupby("a").agg_list("b").to_pydict()
    expected = [[0, 1, 4], [2, 6], [3, 5]]
    for lt in expected:
        assert lt in res["b"]


def test_agg_concat_on_string() -> None:
    df3 = from_pydict({"a": ["the", " quick", " brown", " fox"]})
    res = df3.agg(col("a").agg_concat()).to_pydict()
    assert res["a"] == ["the quick brown fox"]


def test_agg_concat_on_string_groupby() -> None:
    df3 = from_pydict({"a": ["the", " quick", " brown", " fox"], "b": [1, 2, 1, 2]})
    res = df3.groupby("b").agg_concat("a").to_pydict()
    expected = ["the brown", " quick fox"]
    for txt in expected:
        assert txt in res["a"]


def test_agg_concat_on_string_null() -> None:
    df3 = from_pydict({"a": ["the", " quick", None, " fox"]})
    res = df3.agg(col("a").agg_concat()).to_pydict()
    expected = ["the quick fox"]
    assert res["a"] == expected


def test_agg_concat_on_string_groupby_null() -> None:
    df3 = from_pydict({"a": ["the", " quick", None, " fox"], "b": [1, 2, 1, 2]})
    res = df3.groupby("b").agg_concat("a").to_pydict()
    expected = ["the", " quick fox"]
    for txt in expected:
        assert txt in res["a"]


def test_agg_concat_on_string_null_list() -> None:
    df3 = from_pydict({"a": [None, None, None, None], "b": [1, 2, 1, 2]}).with_column(
        "a", col("a").cast(DataType.string())
    )
    res = df3.agg(col("a").agg_concat()).to_pydict()
    expected = [None]
    assert res["a"] == expected
    assert len(res["a"]) == 1


def test_agg_concat_on_string_groupby_null_list() -> None:
    df3 = from_pydict({"a": [None, None, None, None], "b": [1, 2, 1, 2]}).with_column(
        "a", col("a").cast(DataType.string())
    )
    res = df3.groupby("b").agg_concat("a").to_pydict()
    expected = [None, None]
    assert res["a"] == expected
    assert len(res["a"]) == len(expected)


@pytest.mark.parametrize(
    "dtype", daft_nonnull_types + daft_null_types, ids=[f"{_}" for _ in daft_nonnull_types + daft_null_types]
)
def test_global_set_aggs(dtype) -> None:
    input = [None, 0, 1, 2, None, 4, 2, 1, None]
    if dtype == DataType.date():
        input = [datetime.date(2020 + x, 1 + x, 1 + x) if x is not None else None for x in input]
    elif dtype == DataType.bool():
        input = [bool(x) if x is not None else None for x in input]
    elif dtype == DataType.string():
        input = [str(x) if x is not None else None for x in input]
    elif dtype == DataType.binary():
        input = [bytes(x) if x is not None else None for x in input]
    elif dtype == DataType.null():
        input = [None for _ in input]
    daft_table = MicroPartition.from_pydict({"input": input})
    daft_table = daft_table.eval_expression_list([col("input").cast(dtype)])

    # Test without nulls
    result = daft_table.eval_expression_list([col("input").alias("set").agg_set()])
    assert result.get_column_by_name("set").datatype() == DataType.list(dtype)
    expected = [x for x in set(input) if x is not None]
    result_set = result.to_pydict()["set"][0]
    _assert_all_hashable(result_set, "test_global_set_aggs")
    # Check length
    assert len(result_set) == len(expected)
    # Convert both to sets to ignore order
    assert set(result_set) == set(expected)


def test_global_pyobj_set_aggs() -> None:
    pytest.skip(reason="Skipping because Python objects are not supported for set aggregation")
    obj1, obj2, obj3 = object(), object(), object()
    input = [obj1, obj2, None, obj3, obj1, None, obj2]
    table = MicroPartition.from_pydict({"input": input})

    # Should panic because Python objects are not implemented
    with pytest.raises(Exception, match="Python not implemented"):
        table.eval_expression_list([col("input").alias("set").agg_set()])


@pytest.mark.parametrize(
    "dtype", daft_nonnull_types + daft_null_types, ids=[f"{_}" for _ in daft_nonnull_types + daft_null_types]
)
def test_grouped_set_aggs(dtype) -> None:
    groups = [1, 2, 3, 1, 2, 3, 1, 2, 3, 1]
    input = [None, 0, 1, 2, 0, 2, None, 1, None, 3]

    if dtype == DataType.date():
        input = [datetime.date(2020 + x, 1 + x, 1 + x) if x is not None else None for x in input]
    elif dtype == DataType.bool():
        input = [bool(x) if x is not None else None for x in input]
    elif dtype == DataType.string():
        input = [str(x) if x is not None else None for x in input]
    elif dtype == DataType.binary():
        input = [bytes(x) if x is not None else None for x in input]
    elif dtype == DataType.null():
        input = [None for _ in input]

    daft_table = MicroPartition.from_pydict({"groups": groups, "input": input})
    daft_table = daft_table.eval_expression_list([col("groups"), col("input").cast(dtype)])
    input_as_dtype = daft_table.get_column_by_name("input").to_pylist()

    result = daft_table.agg([col("input").alias("set").agg_set()], group_by=[col("groups")]).sort([col("groups")])
    assert result.get_column_by_name("set").datatype() == DataType.list(dtype)

    result_dict = result.to_pydict()
    assert sorted(result_dict["groups"]) == [1, 2, 3]

    for i, group_set in enumerate(result_dict["set"]):
        _assert_all_hashable(group_set, f"test_grouped_set_aggs (group {result_dict['groups'][i]})")

    group1_set = set(result_dict["set"][0])
    group2_set = set(result_dict["set"][1])
    group3_set = set(result_dict["set"][2])

    group1_expected = {input_as_dtype[i] for i in [3, 9]}
    group2_expected = {input_as_dtype[i] for i in [1, 4, 7]}
    group3_expected = {input_as_dtype[i] for i in [2, 5]}

    if dtype == DataType.null():
        group1_expected = set()
        group2_expected = set()
        group3_expected = set()

    assert group1_set == group1_expected, f"Group 1 set incorrect. Expected {group1_expected}, got {group1_set}"
    assert group2_set == group2_expected, f"Group 2 set incorrect. Expected {group2_expected}, got {group2_set}"
    assert group3_set == group3_expected, f"Group 3 set incorrect. Expected {group3_expected}, got {group3_set}"
    assert None not in group1_set and None not in group2_set and None not in group3_set


def test_grouped_pyobj_set_aggs() -> None:
    pytest.skip(reason="Skipping because Python objects are not supported for set aggregation")
    obj1, obj2, obj3 = object(), object(), object()
    groups = [1, 2, 1, 2, 1, 2]
    input = [obj1, obj2, None, obj3, obj1, None]

    table = MicroPartition.from_pydict({"groups": groups, "input": input})

    # Should error because Python objects are not hashable
    with pytest.raises(ValueError, match="Cannot perform set aggregation on elements that are not hashable"):
        table.agg([col("input").alias("set").agg_set()], group_by=[col("groups")])


def test_grouped_list_set_aggs() -> None:
    pytest.skip(reason="Skipping because list set aggregation is not yet implemented")
    groups = [None, 1, None, 1, 2, 2, 1]
    input = [[1], [2, 3, 4], [5, None], None, [], [8, 9], [2, 3, 4]]  # Added duplicate list
    expected_idx = [[1, 3, 6], [4, 5], [0, 2]]

    daft_table = MicroPartition.from_pydict({"groups": groups, "input": input})
    daft_table = daft_table.eval_expression_list([col("groups"), col("input")])

    # Test without nulls
    result = daft_table.agg([col("input").alias("set").agg_set()], group_by=[col("groups")]).sort([col("groups")])
    assert result.get_column_by_name("set").datatype() == DataType.list(DataType.list(DataType.int64()))
    input_as_dtype = daft_table.get_column_by_name("input").to_pylist()

    # Convert lists to tuples for hashing
    def to_hashable(lst):
        return tuple(lst) if lst is not None else None

    expected_groups_no_nulls = []
    for group in expected_idx:
        group_values = [input_as_dtype[i] for i in group if input_as_dtype[i] is not None]
        # First collect unique values
        unique_values = set(to_hashable(x) for x in group_values)
        # Sort non-None values only
        sorted_values = sorted(v for v in unique_values if v is not None)
        # Convert back to lists for comparison
        expected_groups_no_nulls.append([list(x) for x in sorted_values])

    assert result.to_pydict() == {"groups": [1, 2, None], "set": expected_groups_no_nulls}


def test_grouped_struct_set_aggs() -> None:
    pytest.skip(reason="Skipping because struct set aggregation is not yet implemented")
    groups = [None, 1, None, 1, 2, 2, 1]
    input = [
        {"x": 1, "y": 2},
        {"x": 3, "y": 4},
        {"x": 5, "y": None},
        None,
        {"x": 6, "y": 7},
        {"x": 8, "y": 9},
        {"x": 3, "y": 4},  # Added duplicate struct
    ]
    expected_idx = [[1, 3, 6], [4, 5], [0, 2]]

    daft_table = MicroPartition.from_pydict({"groups": groups, "input": input})
    daft_table = daft_table.eval_expression_list([col("groups"), col("input")])

    # Test without nulls
    result = daft_table.agg([col("input").alias("set").agg_set()], group_by=[col("groups")]).sort([col("groups")])
    assert result.get_column_by_name("set").datatype() == DataType.list(
        DataType.struct({"x": DataType.int64(), "y": DataType.int64()})
    )
    input_as_dtype = daft_table.get_column_by_name("input").to_pylist()

    # Convert dicts to tuples for hashing
    def to_hashable(d):
        return tuple(sorted(d.items())) if d is not None else None

    expected_groups_no_nulls = []
    for group in expected_idx:
        group_values = [input_as_dtype[i] for i in group if input_as_dtype[i] is not None]
        # First collect unique values
        unique_values = set(to_hashable(x) for x in group_values)
        # Sort non-None values only
        sorted_values = sorted(v for v in unique_values if v is not None)
        # Convert back to dicts for comparison
        expected_groups_no_nulls.append([dict(x) for x in sorted_values])

    assert result.to_pydict() == {"groups": [1, 2, None], "set": expected_groups_no_nulls}


def test_set_aggs_empty() -> None:
    daft_table = MicroPartition.from_pydict({"col_A": [], "col_B": []})

    # Test without nulls
    result = daft_table.agg(
        [col("col_A").cast(DataType.int32()).alias("set").agg_set()],
        group_by=[col("col_B")],
    )
    assert result.get_column_by_name("set").datatype() == DataType.list(DataType.int32())
    assert result.to_pydict() == {"col_B": [], "set": []}


test_table_bool_agg_cases = [
    ([], {"bool_and": [None], "bool_or": [None]}),
    ([None], {"bool_and": [None], "bool_or": [None]}),
    ([None, None, None], {"bool_and": [None], "bool_or": [None]}),
    ([True], {"bool_and": [True], "bool_or": [True]}),
    ([False], {"bool_and": [False], "bool_or": [False]}),
    ([True, True], {"bool_and": [True], "bool_or": [True]}),
    ([False, False], {"bool_and": [False], "bool_or": [False]}),
    ([True, False], {"bool_and": [False], "bool_or": [True]}),
    ([None, True], {"bool_and": [True], "bool_or": [True]}),
    ([None, False], {"bool_and": [False], "bool_or": [False]}),
    ([True, None, True], {"bool_and": [True], "bool_or": [True]}),
    ([False, None, False], {"bool_and": [False], "bool_or": [False]}),
    ([True, None, False], {"bool_and": [False], "bool_or": [True]}),
]


@pytest.mark.parametrize("case", test_table_bool_agg_cases, ids=[f"{_}" for _ in test_table_bool_agg_cases])
def test_table_bool_agg(case) -> None:
    input, expected = case
    daft_table = MicroPartition.from_pydict({"input": input})
    daft_table = daft_table.eval_expression_list([col("input").cast(DataType.bool())])
    daft_table = daft_table.eval_expression_list(
        [
            col("input").alias("bool_and").bool_and(),
            col("input").alias("bool_or").bool_or(),
        ]
    )

    res = daft_table.to_pydict()
    assert res == expected



from __future__ import annotations

from datetime import datetime

import numpy as np
import pandas as pd
import pytest

from daft.datatype import DataType
from daft.expressions import col
from daft.udf import udf
from tests.conftest import assert_df_equals
from tests.dataframe.test_aggregations import _assert_all_hashable


def test_sum(daft_df, service_requests_csv_pd_df, repartition_nparts, with_morsel_size):
    """Sums across an entire column for the entire table."""
    daft_df = daft_df.repartition(repartition_nparts).sum(col("Unique Key").alias("unique_key_sum"))
    service_requests_csv_pd_df = pd.DataFrame.from_records(
        [{"unique_key_sum": service_requests_csv_pd_df["Unique Key"].sum()}]
    )
    daft_pd_df = daft_df.to_pandas()
    assert_df_equals(daft_pd_df, service_requests_csv_pd_df, sort_key="unique_key_sum")


def test_approx_percentiles(daft_df, service_requests_csv_pd_df, repartition_nparts, with_morsel_size):
    """Computes approx percentile across an entire column for the entire table."""
    daft_df = daft_df.repartition(repartition_nparts).agg(
        col("Unique Key").alias("unique_key_median").approx_percentiles([0.25, 0.5, 0.75])
    )
    service_requests_csv_pd_df = pd.DataFrame.from_records(
        [{"unique_key_median": [service_requests_csv_pd_df["Unique Key"].quantile(p) for p in (0.25, 0.5, 0.75)]}]
    )
    daft_pd_df = daft_df.to_pandas()
    # Assert approximate median to be at 2% of exact median
    pd.testing.assert_series_equal(
        daft_pd_df["unique_key_median"], service_requests_csv_pd_df["unique_key_median"], check_exact=False, rtol=0.02
    )


def test_mean(daft_df, service_requests_csv_pd_df, repartition_nparts, with_morsel_size):
    """Averages across a column for entire table."""
    daft_df = daft_df.repartition(repartition_nparts).mean(col("Unique Key").alias("unique_key_mean"))
    service_requests_csv_pd_df = pd.DataFrame.from_records(
        [{"unique_key_mean": service_requests_csv_pd_df["Unique Key"].mean()}]
    )
    daft_pd_df = daft_df.to_pandas()
    assert_df_equals(daft_pd_df, service_requests_csv_pd_df, sort_key="unique_key_mean")


def test_min(daft_df, service_requests_csv_pd_df, repartition_nparts, with_morsel_size):
    """Min across a column for entire table."""
    daft_df = daft_df.repartition(repartition_nparts).min(col("Unique Key").alias("unique_key_min"))
    service_requests_csv_pd_df = pd.DataFrame.from_records(
        [{"unique_key_min": service_requests_csv_pd_df["Unique Key"].min()}]
    )
    daft_pd_df = daft_df.to_pandas()
    assert_df_equals(daft_pd_df, service_requests_csv_pd_df, sort_key="unique_key_min")


def test_max(daft_df, service_requests_csv_pd_df, repartition_nparts, with_morsel_size):
    """Max across a column for entire table."""
    daft_df = daft_df.repartition(repartition_nparts).max(col("Unique Key").alias("unique_key_max"))
    service_requests_csv_pd_df = pd.DataFrame.from_records(
        [{"unique_key_max": service_requests_csv_pd_df["Unique Key"].max()}]
    )
    daft_pd_df = daft_df.to_pandas()
    assert_df_equals(daft_pd_df, service_requests_csv_pd_df, sort_key="unique_key_max")


def test_count(daft_df, service_requests_csv_pd_df, repartition_nparts, with_morsel_size):
    """Count a column for entire table."""
    daft_df = daft_df.repartition(repartition_nparts).count(col("Unique Key").alias("unique_key_count"))
    service_requests_csv_pd_df = pd.DataFrame.from_records(
        [{"unique_key_count": service_requests_csv_pd_df["Unique Key"].count()}]
    )
    service_requests_csv_pd_df["unique_key_count"] = service_requests_csv_pd_df["unique_key_count"].astype("uint64")
    daft_pd_df = daft_df.to_pandas()
    assert_df_equals(daft_pd_df, service_requests_csv_pd_df, sort_key="unique_key_count")


def test_list(daft_df, service_requests_csv_pd_df, repartition_nparts, with_morsel_size):
    """List agg a column for entire table."""
    daft_df = daft_df.repartition(repartition_nparts).agg_list(col("Unique Key").alias("unique_key_list")).collect()
    unique_key_list = service_requests_csv_pd_df["Unique Key"].to_list()

    result_list = daft_df.to_pydict()["unique_key_list"]
    assert len(result_list) == 1
    assert set(result_list[0]) == set(unique_key_list)


@pytest.mark.parametrize("npartitions", [1, 2])
def test_set(make_df, npartitions, with_morsel_size):
    df = make_df(
        {
            "values": [1, 2, 2, 3, 3, 3],
        },
        repartition=npartitions,
    )
    df = df.agg([col("values").agg_set().alias("set")])
    df.collect()
    result = df.to_pydict()["set"][0]
    _assert_all_hashable(result, "test_set")
    assert set(result) == {1, 2, 3}


def test_global_agg(daft_df, service_requests_csv_pd_df, repartition_nparts, with_morsel_size):
    """Averages across a column for entire table."""
    daft_df = daft_df.repartition(repartition_nparts).agg(
        [
            col("Unique Key").mean().alias("unique_key_mean"),
            col("Unique Key").sum().alias("unique_key_sum"),
            col("Borough").min().alias("borough_min"),
            col("Borough").max().alias("borough_max"),
        ]
    )
    service_requests_csv_pd_df = pd.DataFrame.from_records(
        [
            {
                "unique_key_mean": service_requests_csv_pd_df["Unique Key"].mean(),
                "unique_key_sum": service_requests_csv_pd_df["Unique Key"].sum(),
                "borough_min": service_requests_csv_pd_df["Borough"].min(),
                "borough_max": service_requests_csv_pd_df["Borough"].max(),
            }
        ]
    )
    daft_pd_df = daft_df.to_pandas()
    assert_df_equals(daft_pd_df, service_requests_csv_pd_df, sort_key="unique_key_mean")


def test_filtered_sum(daft_df, service_requests_csv_pd_df, repartition_nparts, with_morsel_size):
    """Sums across an entire column for the entire table filtered by a certain condition."""
    daft_df = (
        daft_df.repartition(repartition_nparts)
        .where(col("Borough") == "BROOKLYN")
        .sum(col("Unique Key").alias("unique_key_sum"))
    )
    service_requests_csv_pd_df = pd.DataFrame.from_records(
        [
            {
                "unique_key_sum": service_requests_csv_pd_df[service_requests_csv_pd_df["Borough"] == "BROOKLYN"][
                    "Unique Key"
                ].sum()
            }
        ]
    )
    daft_pd_df = daft_df.to_pandas()
    assert_df_equals(daft_pd_df, service_requests_csv_pd_df, sort_key="unique_key_sum")


@pytest.mark.parametrize(
    "keys",
    [
        pytest.param(["Borough"], id="NumGroupByKeys:1"),
        pytest.param(["Borough", "Complaint Type"], id="NumGroupByKeys:2"),
    ],
)
def test_sum_groupby(daft_df, service_requests_csv_pd_df, repartition_nparts, keys, with_morsel_size):
    """Sums across groups."""
    daft_df = daft_df.repartition(repartition_nparts).groupby(*[col(k) for k in keys]).sum(col("Unique Key"))
    service_requests_csv_pd_df = service_requests_csv_pd_df.groupby(keys).sum("Unique Key").reset_index()
    daft_pd_df = daft_df.to_pandas()
    assert_df_equals(daft_pd_df, service_requests_csv_pd_df, sort_key=keys)


@pytest.mark.parametrize(
    "keys",
    [
        pytest.param(["Borough"], id="NumGroupByKeys:1"),
        pytest.param(["Borough", "Complaint Type"], id="NumGroupByKeys:2"),
    ],
)
def test_approx_percentile_groupby(daft_df, service_requests_csv_pd_df, repartition_nparts, keys, with_morsel_size):
    """Computes approx percentile across groups."""
    daft_df = (
        daft_df.repartition(repartition_nparts)
        .groupby(*[col(k) for k in keys])
        .agg(col("Unique Key").approx_percentiles([0.25, 0.5, 0.75]))
    )
    service_requests_csv_pd_df = (
        service_requests_csv_pd_df.groupby(keys)["Unique Key"]
        .agg(lambda x: list(np.percentile(x, [25, 50, 75])))
        .reset_index()
    )
    daft_pd_df = daft_df.to_pandas()
    # Assert approximate percentiles to be at 2% of exact percentiles
    pd.testing.assert_series_equal(
        daft_pd_df["Unique Key"], service_requests_csv_pd_df["Unique Key"], check_exact=False, rtol=0.02
    )


@pytest.mark.parametrize(
    "keys",
    [
        pytest.param(["Borough"], id="NumGroupByKeys:1"),
        pytest.param(["Borough", "Complaint Type"], id="NumGroupByKeys:2"),
    ],
)
def test_mean_groupby(daft_df, service_requests_csv_pd_df, repartition_nparts, keys, with_morsel_size):
    """Sums across groups."""
    daft_df = daft_df.repartition(repartition_nparts).groupby(*[col(k) for k in keys]).mean(col("Unique Key"))
    service_requests_csv_pd_df = service_requests_csv_pd_df.groupby(keys).mean("Unique Key").reset_index()
    daft_pd_df = daft_df.to_pandas()
    assert_df_equals(daft_pd_df, service_requests_csv_pd_df, sort_key=keys)


@pytest.mark.parametrize(
    "keys",
    [
        pytest.param(["Borough"], id="NumGroupByKeys:1"),
        pytest.param(["Borough", "Complaint Type"], id="NumGroupByKeys:2"),
    ],
)
def test_count_groupby(daft_df, service_requests_csv_pd_df, repartition_nparts, keys, with_morsel_size):
    """Count across groups."""
    daft_df = daft_df.repartition(repartition_nparts).groupby(*[col(k) for k in keys]).count()
    service_requests_csv_pd_df = service_requests_csv_pd_df.groupby(keys).count().reset_index()
    for cname in service_requests_csv_pd_df:
        if cname not in keys:
            service_requests_csv_pd_df[cname] = service_requests_csv_pd_df[cname].astype("uint64")
    daft_pd_df = daft_df.to_pandas()
    assert_df_equals(daft_pd_df, service_requests_csv_pd_df, sort_key=keys)


@pytest.mark.parametrize(
    "keys",
    [
        pytest.param(["Borough"], id="NumGroupByKeys:1"),
        pytest.param(["Borough", "Complaint Type"], id="NumGroupByKeys:2"),
    ],
)
def test_min_groupby(daft_df, service_requests_csv_pd_df, repartition_nparts, keys, with_morsel_size):
    """Min across groups."""
    daft_df = (
        daft_df.repartition(repartition_nparts)
        .groupby(*[col(k) for k in keys])
        .min(col("Unique Key"), col("Created Date"))
    )
    service_requests_csv_pd_df = (
        service_requests_csv_pd_df.groupby(keys)[["Unique Key", "Created Date"]].min().reset_index()
    )
    daft_pd_df = daft_df.to_pandas()
    assert_df_equals(daft_pd_df, service_requests_csv_pd_df, sort_key=keys)


@pytest.mark.parametrize(
    "keys",
    [
        pytest.param(["Borough"], id="NumGroupByKeys:1"),
        pytest.param(["Borough", "Complaint Type"], id="NumGroupByKeys:2"),
    ],
)
def test_max_groupby(daft_df, service_requests_csv_pd_df, repartition_nparts, keys, with_morsel_size):
    """Max across groups."""
    daft_df = (
        daft_df.repartition(repartition_nparts)
        .groupby(*[col(k) for k in keys])
        .max(col("Unique Key"), col("Created Date"))
    )
    service_requests_csv_pd_df = (
        service_requests_csv_pd_df.groupby(keys)[["Unique Key", "Created Date"]].max().reset_index()
    )
    daft_pd_df = daft_df.to_pandas()
    assert_df_equals(daft_pd_df, service_requests_csv_pd_df, sort_key=keys)


@pytest.mark.parametrize(
    "keys",
    [
        pytest.param(["Borough"], id="NumGroupSortKeys:1"),
        pytest.param(["Borough", "Complaint Type"], id="NumGroupSortKeys:2"),
    ],
)
def test_sum_groupby_sorted(daft_df, service_requests_csv_pd_df, repartition_nparts, keys, with_morsel_size):
    """Test sorting after a groupby."""
    daft_df = (
        daft_df.repartition(repartition_nparts)
        .groupby(*[col(k) for k in keys])
        .sum(col("Unique Key"))
        .sort(by=[col(k) for k in keys], desc=True)
    )
    service_requests_csv_pd_df = (
        service_requests_csv_pd_df.groupby(keys).sum("Unique Key").sort_values(by=keys, ascending=False)
    ).reset_index()
    daft_pd_df = daft_df.to_pandas()
    assert_df_equals(daft_pd_df, service_requests_csv_pd_df, assert_ordering=True)


@pytest.mark.parametrize(
    "keys",
    [
        pytest.param(["Borough"], id="NumGroupSortKeys:1"),
        pytest.param(["Borough", "Complaint Type"], id="NumGroupSortKeys:2"),
    ],
)
def test_map_groups(daft_df, service_requests_csv_pd_df, repartition_nparts, keys, with_morsel_size):
    """Test map_groups."""

    @udf(return_dtype=DataType.float64())
    def average_resolution_time(created_date, closed_date):
        # Calculate the time difference in seconds between created and closed date
        times = [
            (
                datetime.strptime(closed, "%m/%d/%Y %I:%M:%S %p") - datetime.strptime(created, "%m/%d/%Y %I:%M:%S %p")
            ).seconds
            for created, closed in zip(created_date.to_pylist(), closed_date.to_pylist())
            if closed
        ]

        # Calculate the average if times list is not empty, else return [None]
        average_time = [sum(times) / len(times)] if times else [None]

        return average_time

    daft_df = (
        daft_df.repartition(repartition_nparts)
        .groupby(*[col(k) for k in keys])
        .map_groups(average_resolution_time(daft_df["Created Date"], daft_df["Closed Date"]))
    )
    daft_df = daft_df.select(*[col(k) for k in keys], col("Created Date").alias("Resolution Time in Seconds"))
    daft_pd_df = daft_df.to_pandas()

    service_requests_csv_pd_df["Created Date"] = pd.to_datetime(service_requests_csv_pd_df["Created Date"])
    service_requests_csv_pd_df["Closed Date"] = pd.to_datetime(service_requests_csv_pd_df["Closed Date"])
    service_requests_csv_pd_df["Resolution Time"] = (
        service_requests_csv_pd_df["Closed Date"] - service_requests_csv_pd_df["Created Date"]
    )
    service_requests_csv_pd_df["Resolution Time in Seconds"] = service_requests_csv_pd_df["Resolution Time"].dt.seconds
    average_resolution_time_pd_df = (
        service_requests_csv_pd_df.groupby(keys)["Resolution Time in Seconds"].mean().reset_index()
    )

    assert_df_equals(daft_pd_df, average_resolution_time_pd_df, sort_key=keys)



from __future__ import annotations

import pyarrow as pa
import pytest

import daft
from daft.datatype import DataType
from daft.expressions import col
from daft.recordbatch import MicroPartition

table = MicroPartition.from_pydict({"a": [[1, 2], [3, 4], [5, None], [None, None], None]})
fixed_dtype = DataType.fixed_size_list(DataType.int64(), 2)
fixed_table = table.eval_expression_list([col("a").cast(fixed_dtype)])


@pytest.mark.parametrize("table", [table, fixed_table])
def test_list_sum(table):
    result = table.eval_expression_list([col("a").list.sum()])
    assert result.to_pydict() == {"a": [3, 7, 5, None, None]}


@pytest.mark.parametrize("table", [table, fixed_table])
def test_list_mean(table):
    result = table.eval_expression_list([col("a").list.mean()])
    assert result.to_pydict() == {"a": [1.5, 3.5, 5, None, None]}


@pytest.mark.parametrize("table", [table, fixed_table])
def test_list_min(table):
    result = table.eval_expression_list([col("a").list.min()])
    assert result.to_pydict() == {"a": [1, 3, 5, None, None]}


@pytest.mark.parametrize("table", [table, fixed_table])
def test_list_max(table):
    result = table.eval_expression_list([col("a").list.max()])
    assert result.to_pydict() == {"a": [2, 4, 5, None, None]}


def test_list_numeric_aggs_empty_table():
    empty_table = MicroPartition.from_pydict(
        {
            "col": pa.array([], type=pa.list_(pa.int64())),
            "fixed_col": pa.array([], type=pa.list_(pa.int64(), 2)),
        }
    )

    result = empty_table.eval_expression_list(
        [
            col("col").cast(DataType.list(DataType.int64())).list.sum().alias("col_sum"),
            col("col").list.mean().alias("col_mean"),
            col("col").list.min().alias("col_min"),
            col("col").list.max().alias("col_max"),
            col("fixed_col").list.sum().alias("fixed_col_sum"),
            col("fixed_col").list.mean().alias("fixed_col_mean"),
            col("fixed_col").list.min().alias("fixed_col_min"),
            col("fixed_col").list.max().alias("fixed_col_max"),
        ]
    )
    assert result.to_pydict() == {
        "col_sum": [],
        "col_mean": [],
        "col_min": [],
        "col_max": [],
        "fixed_col_sum": [],
        "fixed_col_mean": [],
        "fixed_col_min": [],
        "fixed_col_max": [],
    }


def test_list_numeric_aggs_with_groupby():
    df = daft.from_pydict(
        {
            "group_col": [1, 1, 1, 1, 2, 2, 2, 2, 3, 3, 3, 3],
            "id_col": [3, 1, 2, 2, 5, 4, None, 3, None, None, None, None],
        }
    )

    # Group by and test aggregates.
    grouped_df = df.groupby("group_col").agg(daft.col("id_col").agg_list().alias("ids_col"))
    result = grouped_df.select(
        col("group_col"),
        col("ids_col").list.sum().alias("ids_col_sum"),
        col("ids_col").list.mean().alias("ids_col_mean"),
        col("ids_col").list.min().alias("ids_col_min"),
        col("ids_col").list.max().alias("ids_col_max"),
    ).sort("group_col", desc=False)
    result_dict = result.to_pydict()
    expected = {
        "group_col": [1, 2, 3],
        "ids_col_sum": [8, 12, None],
        "ids_col_mean": [2.0, 4.0, None],
        "ids_col_min": [1, 3, None],
        "ids_col_max": [3, 5, None],
    }
    assert result_dict == expected

    # Cast to fixed size list, group by, and test aggregates.
    grouped_df = grouped_df.with_column("ids_col", col("ids_col").cast(DataType.fixed_size_list(DataType.int64(), 4)))
    result = grouped_df.select(
        col("group_col"),
        col("ids_col").list.sum().alias("ids_col_sum"),
        col("ids_col").list.mean().alias("ids_col_mean"),
        col("ids_col").list.min().alias("ids_col_min"),
        col("ids_col").list.max().alias("ids_col_max"),
    ).sort("group_col", desc=False)
    result_dict = result.to_pydict()
    assert result_dict == expected



from __future__ import annotations

from typing import Any

import dask.dataframe as dd
import numpy as np
import pandas as pd
import pytest

import daft
from daft import DataType
from tests.conftest import get_tests_daft_runner_name


class MyObj:
    def __init__(self, x: int):
        self._x = x

    def __eq__(self, other: Any) -> bool:
        return isinstance(other, MyObj) and self._x == other._x


DATA = {
    "intcol": [1, 2, 3],
    "strcol": ["a", "b", "c"],
}


@pytest.mark.skipif(get_tests_daft_runner_name() != "ray", reason="Needs to run on Ray runner")
@pytest.mark.parametrize("n_partitions", [1, 2])
def test_to_dask_dataframe_all_arrow(n_partitions: int):
    df = daft.from_pydict(DATA).repartition(n_partitions)
    df = df.with_column("floatcol", df["intcol"].cast(DataType.float64()))
    ddf = df.to_dask_dataframe()

    rows = sorted(ddf.compute().to_dict("records"), key=lambda r: r["intcol"])
    assert rows == sorted(
        [
            {"intcol": 1, "strcol": "a", "floatcol": 1.0},
            {"intcol": 2, "strcol": "b", "floatcol": 2.0},
            {"intcol": 3, "strcol": "c", "floatcol": 3.0},
        ],
        key=lambda r: r["intcol"],
    )


@pytest.mark.skipif(get_tests_daft_runner_name() != "ray", reason="Needs to run on Ray runner")
@pytest.mark.parametrize("n_partitions", [1, 2])
def test_to_dask_dataframe_all_arrow_with_schema(n_partitions: int):
    df = daft.from_pydict(DATA).repartition(n_partitions)
    df = df.with_column("floatcol", df["intcol"].cast(DataType.float64()))
    ddf = df.to_dask_dataframe({"intcol": np.int64, "strcol": np.str_, "floatcol": np.float64})

    rows = sorted(ddf.compute().to_dict("records"), key=lambda r: r["intcol"])
    assert rows == sorted(
        [
            {"intcol": 1, "strcol": "a", "floatcol": 1.0},
            {"intcol": 2, "strcol": "b", "floatcol": 2.0},
            {"intcol": 3, "strcol": "c", "floatcol": 3.0},
        ],
        key=lambda r: r["intcol"],
    )


@pytest.mark.skipif(get_tests_daft_runner_name() != "ray", reason="Needs to run on Ray runner")
@pytest.mark.parametrize("n_partitions", [1, 2])
def test_to_dask_dataframe_with_py(n_partitions: int):
    df = daft.from_pydict(DATA).repartition(n_partitions)
    df = df.with_column("pycol", df["intcol"].apply(lambda x: MyObj(x), DataType.python()))
    ddf = df.to_dask_dataframe()

    rows = sorted(ddf.compute().to_dict("records"), key=lambda r: r["intcol"])
    assert rows == sorted(
        [
            {"intcol": 1, "strcol": "a", "pycol": MyObj(1)},
            {"intcol": 2, "strcol": "b", "pycol": MyObj(2)},
            {"intcol": 3, "strcol": "c", "pycol": MyObj(3)},
        ],
        key=lambda r: r["intcol"],
    )


@pytest.mark.skipif(get_tests_daft_runner_name() != "ray", reason="Needs to run on Ray runner")
@pytest.mark.parametrize("n_partitions", [1, 2])
def test_to_dask_dataframe_with_numpy(n_partitions: int):
    df = daft.from_pydict(DATA).repartition(n_partitions)
    df = df.with_column("npcol", df["intcol"].apply(lambda _: np.ones((3, 3)), DataType.python()))
    ddf = df.to_dask_dataframe()

    rows = sorted(ddf.compute().to_dict("records"), key=lambda r: r["intcol"])
    np.testing.assert_equal(
        rows,
        sorted(
            [
                {"intcol": 1, "strcol": "a", "npcol": np.ones((3, 3))},
                {"intcol": 2, "strcol": "b", "npcol": np.ones((3, 3))},
                {"intcol": 3, "strcol": "c", "npcol": np.ones((3, 3))},
            ],
            key=lambda r: r["intcol"],
        ),
    )


@pytest.mark.skipif(get_tests_daft_runner_name() != "ray", reason="Needs to run on Ray runner")
@pytest.mark.parametrize("n_partitions", [1, 2])
def test_to_dask_dataframe_with_numpy_variable_shaped(n_partitions: int):
    df = daft.from_pydict(DATA).repartition(n_partitions)
    df = df.with_column("npcol", df["intcol"].apply(lambda x: np.ones((x, 3)), DataType.python()))
    ddf = df.to_dask_dataframe()

    rows = sorted(ddf.compute().to_dict("records"), key=lambda r: r["intcol"])
    np.testing.assert_equal(
        rows,
        sorted(
            [
                {"intcol": 1, "strcol": "a", "npcol": np.ones((1, 3))},
                {"intcol": 2, "strcol": "b", "npcol": np.ones((2, 3))},
                {"intcol": 3, "strcol": "c", "npcol": np.ones((3, 3))},
            ],
            key=lambda r: r["intcol"],
        ),
    )


@pytest.mark.skipif(get_tests_daft_runner_name() != "ray", reason="Needs to run on Ray runner")
@pytest.mark.parametrize("n_partitions", [1, 2])
def test_from_dask_dataframe_all_arrow(n_partitions: int):
    df = pd.DataFrame(DATA)
    df["floatcol"] = df["intcol"].astype(float)
    ddf = dd.from_pandas(df, npartitions=n_partitions)

    daft_df = daft.from_dask_dataframe(ddf)
    out_df = daft_df.to_pandas()
    pd.testing.assert_frame_equal(out_df, df)


# @pytest.mark.skipif(get_tests_daft_runner_name() != "ray", reason="Needs to run on Ray runner")
@pytest.mark.skip()  # dask doesn't seem to work with object types anymore
@pytest.mark.parametrize("n_partitions", [1, 2])
def test_from_dask_dataframe_tensor(n_partitions: int):
    df = pd.DataFrame(DATA)
    df["tensor"] = pd.Series([np.ones((2, 2)) for _ in range(len(df))], dtype=object)
    ddf = dd.from_pandas(df, npartitions=n_partitions)

    daft_df = daft.from_dask_dataframe(ddf)
    out_df = daft_df.to_pandas()
    pd.testing.assert_frame_equal(out_df, df)


@pytest.mark.skipif(get_tests_daft_runner_name() != "ray", reason="Needs to run on Ray runner")
@pytest.mark.parametrize("n_partitions", [1, 2])
def test_from_dask_dataframe_preview(n_partitions: int):
    df = pd.DataFrame(DATA)
    ddf = dd.from_pandas(df, npartitions=n_partitions)

    daft_df = daft.from_dask_dataframe(ddf)
    assert len(daft_df) == 3
    assert len(daft_df._preview.partition) == 3


@pytest.mark.skipif(get_tests_daft_runner_name() != "ray", reason="Needs to run on Ray runner")
@pytest.mark.parametrize("n_partitions", [1, 2])
def test_from_dask_dataframe_data_longer_than_preview(n_partitions: int):
    df = pd.DataFrame(
        {
            "intcol": [1, 2, 3, 4, 5, 6, 7, 8, 9, 10],
            "strcol": ["a", "b", "c", "d", "e", "f", "g", "h", "i", "j"],
        }
    )
    ddf = dd.from_pandas(df, npartitions=n_partitions)

    daft_df = daft.from_dask_dataframe(ddf)
    assert len(daft_df) == 10
    assert len(daft_df._preview.partition) == 8



from __future__ import annotations

import datetime
import decimal

import pyarrow as pa
import pytest

import daft
from daft import DataType, TimeUnit
from tests.conftest import get_tests_daft_runner_name

PYARROW_GE_11_0_0 = tuple(int(s) for s in pa.__version__.split(".") if s.isnumeric()) >= (11, 0, 0)


@pytest.mark.skipif(
    get_tests_daft_runner_name() != "native", reason="JSON writes are only implemented in the native runner"
)
@pytest.mark.parametrize(
    ["data", "pa_type", "expected_dtype", "expected_inferred_dtype"],
    [
        ([1, 2, None], pa.int64(), DataType.int64(), DataType.int64()),
        (["a", "b", ""], pa.large_string(), DataType.string(), DataType.string()),
        # TODO(desmond): Arrow-rs writes binaries as strings of hexadecimals. Technically the JSON spec (RFC 8259) does not
        #                support BINARY. In practice, libraries such as SIMD-JSON expects binary to be represented as a
        #                string that's properly encoded e.g. you can encode your binary values in base64.
        #                We should make our readers and writers compatible.
        # ([b"a", b"b", b""], pa.large_binary(), DataType.binary(), DataType.string()),
        ([True, False, None], pa.bool_(), DataType.bool(), DataType.bool()),
        ([None, None, None], pa.null(), DataType.null(), DataType.null()),
        (
            [decimal.Decimal("1.23"), decimal.Decimal("1.24"), None],
            pa.decimal128(16, 8),
            DataType.decimal128(16, 8),
            DataType.float64(),
        ),
        ([datetime.date(1994, 1, 1), datetime.date(1995, 1, 1), None], pa.date32(), DataType.date(), DataType.date()),
        (
            [datetime.time(1, 2, 3, 4), datetime.time(5, 6, 7, 8), None],
            pa.time64("us"),
            DataType.time(TimeUnit.us()),
            DataType.time(TimeUnit.us()),
        ),
        (
            [datetime.time(1, 2, 3, 4), datetime.time(5, 6, 7, 8), None],
            pa.time64("ns"),
            DataType.time(TimeUnit.ns()),
            DataType.time(TimeUnit.us()),
        ),
        (
            [datetime.datetime(1994, 1, 1), datetime.datetime(1995, 1, 1), None],
            pa.timestamp("ms"),
            DataType.timestamp(TimeUnit.ms()),
            # NOTE: Seems like the inferred type is seconds because it's written with seconds resolution
            DataType.timestamp(TimeUnit.s()),
        ),
        (
            [datetime.date(1994, 1, 1), datetime.date(1995, 1, 1), None],
            pa.date64(),
            DataType.timestamp(TimeUnit.ms()),
            DataType.timestamp(TimeUnit.s()),
        ),
        # TODO(desmond): Arrow-rs also currently writes durations in ISO 8601 duration format while our reader expects an i64.
        #                We should make our readers and writers compatible.
        # (
        #     [datetime.timedelta(days=1), datetime.timedelta(days=2), None],
        #     pa.duration("ms"),
        #     DataType.duration(TimeUnit.ms()),
        #     # NOTE: Duration ends up being written as int64
        #     DataType.int64(),
        # ),
    ],
)
def test_roundtrip_simple_arrow_types(tmp_path, data, pa_type, expected_dtype, expected_inferred_dtype):
    before = daft.from_arrow(pa.table({"id": pa.array(range(3)), "foo": pa.array(data, type=pa_type)}))
    before = before.concat(before)
    before.write_json(str(tmp_path))
    after = daft.read_json(str(tmp_path))
    assert before.schema()["foo"].dtype == expected_dtype
    assert after.schema()["foo"].dtype == expected_inferred_dtype
    assert before.to_arrow() == after.with_column("foo", after["foo"].cast(expected_dtype)).to_arrow()


@pytest.mark.skipif(
    get_tests_daft_runner_name() != "native", reason="JSON writes are only implemented in the native runner"
)
def test_roundtrip_struct_types(tmp_path):
    struct_data = [
        {"name": "Alice", "age": 30, "city": "New York"},
        {"name": "Bob", "age": 25, "city": "San Francisco"},
        None,
    ]

    pa_struct_type = pa.struct({"name": pa.string(), "age": pa.int64(), "city": pa.string()})

    expected_dtype = DataType.struct({"name": DataType.string(), "age": DataType.int64(), "city": DataType.string()})

    before = daft.from_arrow(pa.table({"id": pa.array(range(3)), "person": pa.array(struct_data, type=pa_struct_type)}))
    before = before.concat(before)
    before.write_json(str(tmp_path))
    after = daft.read_json(str(tmp_path))

    assert before.schema()["person"].dtype == expected_dtype
    assert after.schema()["person"].dtype == expected_dtype
    assert before.to_arrow() == after.to_arrow()


@pytest.mark.skipif(
    get_tests_daft_runner_name() != "native" or not PYARROW_GE_11_0_0,
    reason="JSON writes are only implemented in the native runner, and map types require PyArrow >= 11.0.0",
)
def test_roundtrip_map_types(tmp_path):
    map_data = [
        {"key1": "value1", "key2": "value2"},
        {"key1": "value3"},
        None,
    ]

    pa_map_type = pa.map_(pa.string(), pa.string())
    expected_dtype = DataType.map(DataType.string(), DataType.string())

    before = daft.from_arrow(pa.table({"id": pa.array(range(3)), "metadata": pa.array(map_data, type=pa_map_type)}))
    before = before.concat(before)
    before.write_json(str(tmp_path))
    after = daft.read_json(str(tmp_path))

    assert before.schema()["metadata"].dtype == expected_dtype
    # In JSON we cannot determine if a type is MAP and not STRUCT.
    # However, since casting from struct to map is not implemented, we'll just verify the data is preserved
    # by checking that the struct representation matches the expected structure.
    assert after.schema()["metadata"].dtype == DataType.struct({"key1": DataType.string(), "key2": DataType.string()})
    # Verify the data length is preserved.
    assert before.count_rows() == after.count_rows()


@pytest.mark.skipif(
    get_tests_daft_runner_name() != "native", reason="JSON writes are only implemented in the native runner"
)
def test_roundtrip_nested_struct_with_arrays(tmp_path):
    """Test JSON roundtrip with nested structs containing arrays."""
    nested_struct_data = [
        {"name": "Alice", "scores": [85, 90, 78], "tags": ["student", "active"]},
        {"name": "Bob", "scores": [92, 88, 95], "tags": ["student"]},
        None,
    ]

    pa_nested_struct_type = pa.struct(
        {"name": pa.string(), "scores": pa.list_(pa.int64()), "tags": pa.list_(pa.string())}
    )

    expected_dtype = DataType.struct(
        {"name": DataType.string(), "scores": DataType.list(DataType.int64()), "tags": DataType.list(DataType.string())}
    )

    before = daft.from_arrow(
        pa.table({"id": pa.array(range(3)), "student": pa.array(nested_struct_data, type=pa_nested_struct_type)})
    )
    before = before.concat(before)
    before.write_json(str(tmp_path))
    after = daft.read_json(str(tmp_path))
    assert before.schema()["student"].dtype == expected_dtype
    assert after.schema()["student"].dtype == expected_dtype
    assert before.to_arrow() == after.to_arrow()


@pytest.mark.skipif(
    get_tests_daft_runner_name() != "native", reason="JSON writes are only implemented in the native runner"
)
def test_throws_error_on_duration_and_binary_types(tmp_path):
    # TODO(desmond): Binary and Duration types currently produce inconsistent behaviours between our readers and writers.
    # Our readers expect BINARY to be encoded as plain text, and DURATION to be encoded with i64. Arrow-rs expects BINARY
    # to be hexadecimal encoded and DURATION to follow the ISO 8601 duration format. Until we reconcile this difference,
    # opt for throwing an error on encountering these types.
    duration_data = [
        datetime.timedelta(days=1),
        datetime.timedelta(days=2),
        None,
    ]

    pa_duration_type = pa.duration("ms")

    before_duration = daft.from_arrow(
        pa.table({"id": pa.array(range(3)), "duration": pa.array(duration_data, type=pa_duration_type)})
    )
    before_duration = before_duration.concat(before_duration)

    # Test that writing duration types throws NotImplementedError
    with pytest.raises(
        daft.exceptions.DaftCoreException,
        match="Not Yet Implemented: JSON writes are not supported with extension, timezone with timestamp, binary, or duration data types",
    ):
        before_duration.write_json(str(tmp_path))

    binary_data = [b"hello", b"world", None]
    pa_binary_type = pa.large_binary()
    before_binary = daft.from_arrow(
        pa.table({"id": pa.array(range(3)), "binary": pa.array(binary_data, type=pa_binary_type)})
    )

    with pytest.raises(
        daft.exceptions.DaftCoreException,
        match="Not Yet Implemented: JSON writes are not supported with extension, timezone with timestamp, binary, or duration data types",
    ):
        before_binary.write_json(str(tmp_path))



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from __future__ import annotations

from pyiceberg.catalog import load_catalog
from pyiceberg.schema import Schema
from pyiceberg.types import FixedType, NestedField, StringType, UUIDType
from pyspark.sql import SparkSession
from pyspark.sql.functions import array, col, current_date, date_add, expr, struct

spark = SparkSession.builder.getOrCreate()

spark.sql(
    """
  CREATE DATABASE IF NOT EXISTS default;
"""
)

schema = Schema(
    NestedField(field_id=1, name="uuid_col", field_type=UUIDType(), required=False),
    NestedField(field_id=2, name="fixed_col", field_type=FixedType(25), required=False),
)

catalog = load_catalog(
    "local",
    **{
        "type": "rest",
        "uri": "http://rest:8181",
        "s3.endpoint": "http://minio:9000",
        "s3.access-key-id": "admin",
        "s3.secret-access-key": "password",
    },
)

catalog.create_table(identifier="default.test_uuid_and_fixed_unpartitioned", schema=schema)

spark.sql(
    """
    INSERT INTO default.test_uuid_and_fixed_unpartitioned VALUES
    ('102cb62f-e6f8-4eb0-9973-d9b012ff0967', CAST('1234567890123456789012345' AS BINARY)),
    ('ec33e4b2-a834-4cc3-8c4a-a1d3bfc2f226', CAST('1231231231231231231231231' AS BINARY)),
    ('639cccce-c9d2-494a-a78c-278ab234f024', CAST('12345678901234567ass12345' AS BINARY)),
    ('c1b0d8e0-0b0e-4b1e-9b0a-0e0b0d0c0a0b', CAST('asdasasdads12312312312111' AS BINARY)),
    ('923dae77-83d6-47cd-b4b0-d383e64ee57e', CAST('qweeqwwqq1231231231231111' AS BINARY));
    """
)

spark.sql(
    """
  CREATE OR REPLACE TABLE default.test_null_nan
  USING iceberg
  AS SELECT
    1            AS idx,
    float('NaN') AS col_numeric
UNION ALL SELECT
    2            AS idx,
    null         AS col_numeric
UNION ALL SELECT
    3            AS idx,
    1            AS col_numeric
"""
)

spark.sql(
    """
  CREATE OR REPLACE TABLE default.test_null_nan_rewritten
  USING iceberg
  AS SELECT * FROM default.test_null_nan
"""
)

spark.sql(
    """
CREATE OR REPLACE TABLE default.test_limit as
  SELECT * LATERAL VIEW explode(ARRAY(1, 2, 3, 4, 5, 6, 7, 8, 9, 10)) AS idx;
"""
)

spark.sql(
    """
CREATE OR REPLACE TABLE default.test_positional_mor_deletes (
    dt     date,
    number integer,
    letter string
)
USING iceberg
TBLPROPERTIES (
    'write.delete.mode'='merge-on-read',
    'write.update.mode'='merge-on-read',
    'write.merge.mode'='merge-on-read',
    'format-version'='2'
);
"""
)

# Partitioning is not really needed, but there is a bug:
# https://github.com/apache/iceberg/pull/7685
spark.sql(
    """
    ALTER TABLE default.test_positional_mor_deletes ADD PARTITION FIELD years(dt) AS dt_years
"""
)

spark.sql(
    """
INSERT INTO default.test_positional_mor_deletes
VALUES
    (CAST('2023-03-01' AS date), 1, 'a'),
    (CAST('2023-03-02' AS date), 2, 'b'),
    (CAST('2023-03-03' AS date), 3, 'c'),
    (CAST('2023-03-04' AS date), 4, 'd'),
    (CAST('2023-03-05' AS date), 5, 'e'),
    (CAST('2023-03-06' AS date), 6, 'f'),
    (CAST('2023-03-07' AS date), 7, 'g'),
    (CAST('2023-03-08' AS date), 8, 'h'),
    (CAST('2023-03-09' AS date), 9, 'i'),
    (CAST('2023-03-10' AS date), 10, 'j'),
    (CAST('2023-03-11' AS date), 11, 'k'),
    (CAST('2023-03-12' AS date), 12, 'l');
"""
)

spark.sql(
    """
ALTER TABLE default.test_positional_mor_deletes CREATE TAG tag_12
    """
)

spark.sql(
    """
ALTER TABLE default.test_positional_mor_deletes CREATE BRANCH without_5
    """
)

spark.sql(
    """
DELETE FROM default.test_positional_mor_deletes.branch_without_5 WHERE number = 5
    """
)


spark.sql(
    """
DELETE FROM default.test_positional_mor_deletes WHERE number = 9
"""
)

spark.sql(
    """
  CREATE OR REPLACE TABLE default.test_positional_mor_double_deletes (
    dt     date,
    number integer,
    letter string
  )
  USING iceberg
  TBLPROPERTIES (
    'write.delete.mode'='merge-on-read',
    'write.update.mode'='merge-on-read',
    'write.merge.mode'='merge-on-read',
    'format-version'='2'
  );
"""
)

# Partitioning is not really needed, but there is a bug:
# https://github.com/apache/iceberg/pull/7685
spark.sql(
    """
    ALTER TABLE default.test_positional_mor_double_deletes ADD PARTITION FIELD years(dt) AS dt_years
"""
)

spark.sql(
    """
INSERT INTO default.test_positional_mor_double_deletes
VALUES
    (CAST('2023-03-01' AS date), 1, 'a'),
    (CAST('2023-03-02' AS date), 2, 'b'),
    (CAST('2023-03-03' AS date), 3, 'c'),
    (CAST('2023-03-04' AS date), 4, 'd'),
    (CAST('2023-03-05' AS date), 5, 'e'),
    (CAST('2023-03-06' AS date), 6, 'f'),
    (CAST('2023-03-07' AS date), 7, 'g'),
    (CAST('2023-03-08' AS date), 8, 'h'),
    (CAST('2023-03-09' AS date), 9, 'i'),
    (CAST('2023-03-10' AS date), 10, 'j'),
    (CAST('2023-03-11' AS date), 11, 'k'),
    (CAST('2023-03-12' AS date), 12, 'l');
"""
)

spark.sql(
    """
    DELETE FROM default.test_positional_mor_double_deletes WHERE number = 9
"""
)

spark.sql(
    """
    DELETE FROM default.test_positional_mor_double_deletes WHERE letter == 'f'
"""
)

all_types_dataframe = (
    spark.range(0, 5, 1, 5)
    .withColumnRenamed("id", "longCol")
    .withColumn("intCol", expr("CAST(longCol AS INT)"))
    .withColumn("floatCol", expr("CAST(longCol AS FLOAT)"))
    .withColumn("doubleCol", expr("CAST(longCol AS DOUBLE)"))
    .withColumn("dateCol", date_add(current_date(), 1))
    .withColumn("timestampCol", expr("TO_TIMESTAMP(dateCol)"))
    .withColumn("stringCol", expr("CAST(dateCol AS STRING)"))
    .withColumn("booleanCol", expr("longCol > 5"))
    .withColumn("binaryCol", expr("CAST(longCol AS BINARY)"))
    .withColumn("byteCol", expr("CAST(longCol AS BYTE)"))
    .withColumn("decimalCol", expr("CAST(longCol AS DECIMAL(10, 2))"))
    .withColumn("shortCol", expr("CAST(longCol AS SHORT)"))
    .withColumn("mapCol", expr("MAP(longCol, decimalCol)"))
    .withColumn("arrayCol", expr("ARRAY(longCol)"))
    .withColumn("structCol", expr("STRUCT(mapCol, arrayCol)"))
)

all_types_dataframe.writeTo("default.test_all_types").tableProperty("format-version", "2").partitionedBy(
    "intCol"
).createOrReplace()

for table_name, partition in [
    ("test_partitioned_by_identity", "ts"),
    ("test_partitioned_by_years", "years(dt)"),
    ("test_partitioned_by_months", "months(dt)"),
    ("test_partitioned_by_days", "days(ts)"),
    ("test_partitioned_by_hours", "hours(ts)"),
    ("test_partitioned_by_truncate", "truncate(1, letter)"),
    ("test_partitioned_by_bucket", "bucket(16, number)"),
]:
    spark.sql(
        f"""
      CREATE OR REPLACE TABLE default.{table_name} (
        dt     date,
        ts     timestamp,
        number integer,
        letter string
      )
      USING iceberg;
    """
    )

    spark.sql(f"ALTER TABLE default.{table_name} ADD PARTITION FIELD {partition}")

    spark.sql(
        f"""
    INSERT INTO default.{table_name}
    VALUES
        (CAST('2022-03-01' AS date), CAST('2022-03-01 01:22:00' AS timestamp), 1, 'a'),
        (CAST('2022-03-02' AS date), CAST('2022-03-02 02:22:00' AS timestamp), 2, 'b'),
        (CAST('2022-03-03' AS date), CAST('2022-03-03 03:22:00' AS timestamp), 3, 'c'),
        (CAST('2022-03-04' AS date), CAST('2022-03-04 04:22:00' AS timestamp), 4, 'd'),
        (CAST('2023-03-05' AS date), CAST('2023-03-05 05:22:00' AS timestamp), 5, 'e'),
        (CAST('2023-03-06' AS date), CAST('2023-03-06 06:22:00' AS timestamp), 6, 'f'),
        (CAST('2023-03-07' AS date), CAST('2023-03-07 07:22:00' AS timestamp), 7, 'g'),
        (CAST('2023-03-08' AS date), CAST('2023-03-08 08:22:00' AS timestamp), 8, 'h'),
        (CAST('2023-03-09' AS date), CAST('2023-03-09 09:22:00' AS timestamp), 9, 'i'),
        (CAST('2023-03-10' AS date), CAST('2023-03-10 10:22:00' AS timestamp), 10, 'j'),
        (CAST('2023-03-11' AS date), CAST('2023-03-11 11:22:00' AS timestamp), 11, 'k'),
        (CAST('2023-03-12' AS date), CAST('2023-03-12 12:22:00' AS timestamp), 12, 'l');
    """
    )

# There is an issue with CREATE OR REPLACE
# https://github.com/apache/iceberg/issues/8756
spark.sql(
    """
DROP TABLE IF EXISTS default.test_table_version
"""
)

spark.sql(
    """
CREATE TABLE default.test_table_version (
    dt     date,
    number integer,
    letter string
)
USING iceberg
TBLPROPERTIES (
    'format-version'='1'
);
"""
)

spark.sql(
    """
CREATE TABLE default.test_table_sanitized_character (
    `letter/abc` string
)
USING iceberg
TBLPROPERTIES (
    'format-version'='1'
);
"""
)

spark.sql(
    """
INSERT INTO default.test_table_sanitized_character
VALUES
    ('123')
"""
)

spark.sql(
    """
  CREATE OR REPLACE TABLE default.test_add_new_column
  USING iceberg
  AS SELECT
        1            AS idx
    UNION ALL SELECT
        2            AS idx
    UNION ALL SELECT
        3            AS idx
"""
)

spark.sql("ALTER TABLE default.test_add_new_column ADD COLUMN name STRING")
spark.sql("INSERT INTO default.test_add_new_column VALUES (3, 'abc'), (4, 'def')")

# In Iceberg the data and schema evolves independently. We can add a column
# that should show up when querying the data, but is not yet represented in a Parquet file

spark.sql(
    """
  CREATE OR REPLACE TABLE default.test_new_column_with_no_data
  USING iceberg
  AS SELECT
        1            AS idx
    UNION ALL SELECT
        2            AS idx
    UNION ALL SELECT
        3            AS idx
"""
)

spark.sql("ALTER TABLE default.test_new_column_with_no_data ADD COLUMN name STRING")


###
# Renaming columns test table
###

renaming_columns_dataframe = (
    spark.range(1, 2, 3)
    .withColumnRenamed("id", "idx")
    .withColumn("data", col("idx") * 10)
    .withColumn("structcol", struct("idx"))
    .withColumn("structcol_oldname", struct("idx"))
    .withColumn("nested_list_struct_col", array("structcol", "structcol", "structcol"))
    .withColumn("deleted_and_then_overwritten_col", col("idx"))
)
renaming_columns_dataframe.writeTo("default.test_table_rename").tableProperty("format-version", "2").createOrReplace()
spark.sql("ALTER TABLE default.test_table_rename RENAME COLUMN idx TO idx_renamed")
spark.sql("ALTER TABLE default.test_table_rename RENAME COLUMN structcol.idx TO idx_renamed")
spark.sql("ALTER TABLE default.test_table_rename RENAME COLUMN structcol_oldname TO structcol_2")

test_table_rename_tbl = catalog.load_table("default.test_table_rename")
with test_table_rename_tbl.update_schema() as txn:
    txn.rename_column("nested_list_struct_col.idx", "idx_renamed")


with test_table_rename_tbl.update_schema() as txn:
    txn.delete_column("deleted_and_then_overwritten_col")
    txn.add_column("deleted_and_then_overwritten_col", StringType())


spark.sql(
    """
  CREATE OR REPLACE TABLE default.test_evolve_partitioning (
      dt     date
  )
  USING iceberg
  PARTITIONED BY (months(dt))
"""
)

spark.sql("INSERT INTO default.test_evolve_partitioning VALUES (CAST('2021-01-01' AS date))")

spark.sql(
    """
    ALTER TABLE default.test_evolve_partitioning
    REPLACE PARTITION FIELD dt_month WITH days(dt)
"""
)

spark.sql("INSERT INTO default.test_evolve_partitioning VALUES (CAST('2021-02-01' AS date))")


###
# Multi-snapshot table
###

spark.sql("""
  CREATE OR REPLACE TABLE default.test_snapshotting
  USING iceberg
  AS SELECT
    1            AS idx,
    float('NaN') AS col_numeric
UNION ALL SELECT
    2            AS idx,
    null         AS col_numeric
UNION ALL SELECT
    3            AS idx,
    1            AS col_numeric
""")

spark.sql("INSERT INTO default.test_snapshotting VALUES (4, 1)")



from __future__ import annotations

import os


def get_asset_dir() -> str:
    dir_path = os.path.dirname(os.path.realpath(__file__))
    return dir_path


ASSET_FOLDER = get_asset_dir()

TPCH_QUERIES = f"{ASSET_FOLDER}/tpch-sqlite-queries"
TPCH_DBGEN_DIR = f"{ASSET_FOLDER}/../../data/tpch-dbgen"



from __future__ import annotations

import datetime
import decimal
import os
import pathlib
import posixpath
import time
import unittest.mock as mock
from collections.abc import Iterator

import boto3
import pyarrow as pa
import pytest
from azure.storage.blob import BlobServiceClient
from pytest_lazyfixture import lazy_fixture

import daft
from daft import DataCatalogTable, DataCatalogType
from daft.io.object_store_options import io_config_to_storage_options


@pytest.fixture(params=[1, 2, 8])
def num_partitions(request) -> int:
    return request.param


@pytest.fixture(
    params=[
        pytest.param((lambda _: None, "a"), id="unpartitioned"),
        pytest.param((lambda i: i, "a"), id="int_partitioned"),
        pytest.param((lambda i: i * 1.5, "b"), id="float_partitioned"),
        pytest.param((lambda i: f"foo_{i}", "c"), id="string_partitioned"),
        # Delta-rs Rust writer doesn't support binary partitioning
        # pytest.param((lambda i: f"foo_{i}".encode(), "d"), id="binary_partitioned"),
        pytest.param(
            (lambda i: datetime.datetime(2024, 2, i + 1), "f"),
            id="timestamp_partitioned",
        ),
        pytest.param((lambda i: datetime.date(2024, 2, i + 1), "g"), id="date_partitioned"),
        pytest.param(
            (lambda i: decimal.Decimal(str(1000 + i) + ".567"), "h"),
            id="decimal_partitioned",
        ),
        pytest.param((lambda i: i if i % 2 == 0 else None, "a"), id="partitioned_with_nulls"),
    ]
)
def partition_generator(request) -> tuple[callable, str]:
    return request.param


@pytest.fixture
def base_table() -> pa.Table:
    return pa.table(
        {
            "a": [1, 2, 3],
            "b": [1.1, 2.2, 3.3],
            "c": ["foo", "bar", "baz"],
            "d": [b"foo", b"bar", b"baz"],
            "e": [True, False, True],
            "f": [
                datetime.datetime(2024, 2, 10),
                datetime.datetime(2024, 2, 11),
                datetime.datetime(2024, 2, 12),
            ],
            "g": [
                datetime.date(2024, 2, 10),
                datetime.date(2024, 2, 11),
                datetime.date(2024, 2, 12),
            ],
            "h": pa.array(
                [
                    decimal.Decimal("1234.567"),
                    decimal.Decimal("1233.456"),
                    decimal.Decimal("1232.345"),
                ],
                type=pa.decimal128(7, 3),
            ),
            "i": [[1, 2, 3], [4, 5, 6], [7, 8, 9]],
            "j": [{"x": 1, "y": False}, {"y": True, "z": "foo"}, {"x": 5, "z": "bar"}],
            "k": pa.array(
                [[("x", 1), ("y", 0)], [("a", 2), ("b", 45)], [("c", 4), ("d", 18)]],
                type=pa.map_(pa.string(), pa.int64()),
            ),
            # TODO(Clark): Wait for more temporal type support in Delta Lake.
            # "l": [
            #     datetime.time(hour=1, minute=2, second=4, microsecond=5),
            #     datetime.time(hour=3, minute=4, second=5, microsecond=6),
            #     datetime.time(hour=4, minute=5, second=6, microsecond=7),
            # ],
            # "m": [
            #     datetime.timedelta(days=1, seconds=2, minutes=5, hours=6, weeks=7),
            #     datetime.timedelta(days=2),
            #     datetime.timedelta(hours=4),
            # ],
        }
    )


@pytest.fixture(scope="session")
def data_dir() -> str:
    return "test_data"


##############################
### Glue-specific fixtures ###
##############################


@pytest.fixture(scope="session")
def glue_database_name() -> str:
    return "glue_db"


@pytest.fixture(scope="session")
def glue_table_name() -> str:
    return "glue_table"


@pytest.fixture(scope="function")
def glue_table(
    aws_server: str,
    glue_database_name: str,
    glue_table_name: str,
    s3_uri: str,
    aws_credentials: dict[str, str],
) -> DataCatalogTable:
    glue = boto3.client(
        "glue",
        region_name="us-west-2",
        use_ssl=False,
        endpoint_url=aws_server,
        aws_access_key_id=aws_credentials["AWS_ACCESS_KEY_ID"],
        aws_secret_access_key=aws_credentials["AWS_SECRET_ACCESS_KEY"],
        aws_session_token=aws_credentials["AWS_SESSION_TOKEN"],
    )
    glue.create_database(DatabaseInput={"Name": glue_database_name})
    glue.create_table(
        DatabaseName=glue_database_name,
        TableInput={
            "Name": glue_table_name,
            "StorageDescriptor": {
                "Location": s3_uri,
            },
        },
    )
    return DataCatalogTable(
        catalog=DataCatalogType.GLUE,
        database_name=glue_database_name,
        table_name=glue_table_name,
    )


##############################
### Unity-specific fixtures ###
##############################


@pytest.fixture(scope="session")
def unity_catalog_id() -> str:
    return "unity_catalog"


@pytest.fixture(scope="session")
def unity_database_name() -> str:
    return "unity_db"


@pytest.fixture(scope="session")
def unity_table_name() -> str:
    return "unity_table"


@pytest.fixture(scope="function")
def unity_table_s3(
    s3_uri: str,
    unity_catalog_id: str,
    unity_database_name: str,
    unity_table_name: str,
) -> Iterator[DataCatalogTable]:
    yield from _unity_table(s3_uri, unity_catalog_id, unity_database_name, unity_table_name)


@pytest.fixture(scope="function")
def unity_table_az(
    az_uri: str,
    unity_catalog_id: str,
    unity_database_name: str,
    unity_table_name: str,
) -> Iterator[DataCatalogTable]:
    yield from _unity_table(az_uri, unity_catalog_id, unity_database_name, unity_table_name)


def _unity_table(
    uri: str,
    unity_catalog_id: str,
    unity_database_name: str,
    unity_table_name: str,
) -> Iterator[DataCatalogTable]:
    from databricks.sdk.service.catalog import TableInfo

    with mock.patch("databricks.sdk.WorkspaceClient") as mock_workspace_client:
        instance = mock_workspace_client.return_value
        instance.tables.get.return_value = TableInfo(storage_location=uri)
        # Set required Databricks environment variables before using SDK.
        # NOTE: This URI won't be used by the Databricks SDK since we mock the WorkspaceClient.
        old_env = os.environ.copy()
        os.environ["DATABRICKS_HOST"] = "http://localhost"
        try:
            yield DataCatalogTable(
                catalog=DataCatalogType.UNITY,
                database_name=unity_database_name,
                table_name=unity_table_name,
                catalog_id=unity_catalog_id,
            )
        finally:
            # Restore old set of environment variables.
            os.environ = old_env


@pytest.fixture(scope="function")
def s3_uri(tmp_path: pathlib.Path, data_dir: str) -> str:
    path = posixpath.join(tmp_path, data_dir).strip("/")
    return "s3://" + path


@pytest.fixture(
    scope="function",
    params=[
        None,
        pytest.param(lazy_fixture("glue_table"), marks=pytest.mark.glue),
        pytest.param(lazy_fixture("unity_table_s3"), marks=pytest.mark.unity),
    ],
)
def s3_path(
    request,
    s3_uri: str,
    aws_server: str,
    aws_credentials: dict[str, str],
    reset_aws: None,
) -> tuple[str, daft.io.IOConfig, DataCatalogTable | None]:
    s3 = boto3.resource(
        "s3",
        region_name="us-west-2",
        use_ssl=False,
        endpoint_url=aws_server,
        aws_access_key_id=aws_credentials["AWS_ACCESS_KEY_ID"],
        aws_secret_access_key=aws_credentials["AWS_SECRET_ACCESS_KEY"],
        aws_session_token=aws_credentials["AWS_SESSION_TOKEN"],
    )
    io_config = daft.io.IOConfig(
        s3=daft.io.S3Config(
            endpoint_url=aws_server,
            region_name="us-west-2",
            key_id=aws_credentials["AWS_ACCESS_KEY_ID"],
            access_key=aws_credentials["AWS_SECRET_ACCESS_KEY"],
            session_token=aws_credentials["AWS_SESSION_TOKEN"],
            use_ssl=False,
        )
    )
    # Create bucket for first element of tmp path.
    bucket = s3.Bucket(s3_uri[5:].split("/")[0])
    bucket.create(CreateBucketConfiguration={"LocationConstraint": "us-west-2"})
    # Bucket will get cleared by reset_s3 fixture, so we don't need to delete it at the end of the test via the
    # typical try-yield-finally block.
    return s3_uri, io_config, request.param


###############################
### Azure-specific fixtures ###
###############################


@pytest.fixture(scope="session")
def az_credentials() -> dict[str, str]:
    return {
        # These are the well-known values
        # https://learn.microsoft.com/en-us/azure/storage/common/storage-use-azurite?tabs=visual-studio#well-known-storage-account-and-key
        "ACCOUNT_NAME": "devstoreaccount1",
        "KEY": "Eby8vdM02xNOcqFlqUwJPLlmEtlCDXJ1OUzFT50uSRZ6IFsuFq2UVErCz4I6tq/K1SZFPTOtr/KBHBeksoGMGw==",
    }


@pytest.fixture(scope="session")
def az_server_ip() -> str:
    return "0.0.0.0"


@pytest.fixture(scope="session")
def az_server_port() -> int:
    return 10000


@pytest.fixture(scope="function")
def az_uri(tmp_path: pathlib.Path, data_dir: str) -> str:
    path = data_dir.strip("/").replace("_", "-")
    return "az://" + path


@pytest.fixture(scope="session")
def az_server(az_server_ip: str, az_server_port: int) -> Iterator[str]:
    docker = pytest.importorskip("docker")
    az_server_url = f"http://{az_server_ip}:{az_server_port}"
    client = docker.from_env()
    azurite = client.containers.run(
        "mcr.microsoft.com/azure-storage/azurite",
        f"azurite-blob --loose --blobHost {az_server_ip}",
        detach=True,
        ports={str(az_server_port): str(az_server_port)},
    )
    for _ in range(100):
        if azurite.status == "running":
            break
        elif azurite.status == "exited":
            output = azurite.logs()
            print(f"Azurite container exited without executing tests: {output}")
            pytest.fail("Cannot start mock Azure Blob Storage server")
        time.sleep(0.1)
        azurite.reload()
    try:
        yield az_server_url
    finally:
        azurite.stop()


@pytest.fixture(
    scope="function",
    params=[
        None,
        pytest.param(lazy_fixture("unity_table_az"), marks=pytest.mark.unity),
    ],
)
def az_path(
    az_uri: str, az_server: str, az_credentials: dict[str, str]
) -> Iterator[tuple[str, daft.io.IOConfig, None]]:
    account_name = az_credentials["ACCOUNT_NAME"]
    key = az_credentials["KEY"]
    endpoint_url = f"{az_server}/{account_name}"
    conn_str = f"DefaultEndpointsProtocol=http;AccountName={account_name};AccountKey={key};BlobEndpoint={endpoint_url};"

    io_config = daft.io.IOConfig(
        azure=daft.io.AzureConfig(
            storage_account=az_credentials["ACCOUNT_NAME"],
            access_key=az_credentials["KEY"],
            endpoint_url=endpoint_url,
            use_ssl=False,
        )
    )
    bbs = BlobServiceClient.from_connection_string(conn_str)
    container = az_uri[5:].split("/")[0]
    bbs.create_container(container)
    try:
        yield az_uri, io_config, None
    finally:
        bbs.delete_container(container)


@pytest.fixture(scope="function")
def local_path(tmp_path: pathlib.Path, data_dir: str) -> tuple[str, None, None]:
    path = os.path.join(tmp_path, data_dir)
    os.mkdir(path)
    return path, None, None


@pytest.fixture(
    scope="function",
    params=[
        pytest.param(lazy_fixture("local_path"), marks=pytest.mark.local),
        pytest.param(lazy_fixture("s3_path"), marks=pytest.mark.s3),
        # Azure tests require starting a Docker container + mock server that (1) requires a dev Docker dependency, and
        # (2) takes 15+ seconds to start on every run, so we current mark it as an integration test.
        pytest.param(lazy_fixture("az_path"), marks=(pytest.mark.az, pytest.mark.integration)),
    ],
)
def cloud_paths(
    request,
) -> tuple[str, daft.io.IOConfig | None, DataCatalogTable | None]:
    return request.param


@pytest.fixture(scope="function")
def deltalake_table(
    cloud_paths,
    base_table: pa.Table,
    num_partitions: int,
    partition_generator: callable,
) -> tuple[str, daft.io.IOConfig | None, dict[str, str], list[pa.Table]]:
    partition_generator, col = partition_generator
    path, io_config, catalog_table = cloud_paths
    storage_options = io_config_to_storage_options(io_config, path) if io_config is not None else None
    parts = []
    for i in range(num_partitions):
        # Generate partition value and add partition column.
        part_value = partition_generator(i)
        part = base_table.append_column("part_idx", pa.array([part_value] * 3, type=base_table.column(col).type))
        parts.append(part)
    table = pa.concat_tables(parts)
    deltalake = pytest.importorskip("deltalake")
    deltalake.write_deltalake(
        path,
        table,
        partition_by="part_idx" if partition_generator(0) is not None else None,
        storage_options=storage_options,
    )
    return path, catalog_table, io_config, parts



from __future__ import annotations

from collections.abc import Iterator
from typing import TYPE_CHECKING

from daft.daft import (
    PyPartitionField,
    PyPushdowns,
    PyRecordBatch,
    ScanTask,
)
from daft.io.pushdowns import Pushdowns, SupportsPushdownFilters
from daft.io.scan import ScanOperator

if TYPE_CHECKING:
    from collections.abc import Iterator

    from daft.io.source import DataSource, DataSourceTask
    from daft.schema import Schema


class _DataSourceShim(ScanOperator):
    """!! INTERNAL ONLY .. SHIM TO REUSE EXISTING BACKED WORK !!"""

    _source: DataSource

    def __init__(self, source: DataSource) -> None:
        self._source = source

    def schema(self) -> Schema:
        return self._source.schema

    def name(self) -> str:
        return self._source.name

    def display_name(self) -> str:
        return f"DataSource({self.name()})"

    def partitioning_keys(self) -> list[PyPartitionField]:
        return [pf._partition_field for pf in self._source.get_partition_fields()]

    def can_absorb_filter(self) -> bool:
        return False

    def can_absorb_limit(self) -> bool:
        return False

    def can_absorb_select(self) -> bool:
        return False

    def multiline_display(self) -> list[str]:
        return [
            self.display_name(),
            f"Schema = {self.schema()}",
        ]

    def to_scan_tasks(self, pushdowns: PyPushdowns) -> Iterator[ScanTask]:
        pds = Pushdowns._from_pypushdowns(pushdowns)
        for task in self._source.get_tasks(pds):
            yield ScanTask.python_factory_func_scan_task(
                module=_get_record_batches.__module__,
                func_name=_get_record_batches.__name__,
                func_args=(task,),
                schema=task.schema._schema,
                num_rows=None,
                size_bytes=None,
                pushdowns=pushdowns,
                stats=None,
            )

    def as_pushdown_filter(self) -> SupportsPushdownFilters | None:
        return None


def _get_record_batches(task: DataSourceTask) -> Iterator[PyRecordBatch]:
    """The task instance has been pickled then sent to this stateless method."""
    yield from (rb._recordbatch for mp in task.get_micro_partitions() for rb in mp.get_record_batches())



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Borrowed and modified from [`ray`](https://github.com/ray-project/ray/blob/master/python/ray/cluster_utils.py).
from __future__ import annotations

import copy
import json
import os
import subprocess
import tempfile
from contextlib import contextmanager

import ray
import yaml


class AutoscalingCluster:
    """Create a local autoscaling cluster for testing."""

    def __init__(
        self,
        head_resources: dict,
        worker_node_types: dict,
        autoscaler_v2: bool = False,
        **config_kwargs,
    ):
        """Create the cluster.

        Args:
            head_resources: resources of the head node, including CPU.
            worker_node_types: autoscaler node types config for worker nodes.
        """
        self._head_resources = head_resources
        self._config = self._generate_config(
            head_resources,
            worker_node_types,
            autoscaler_v2=autoscaler_v2,
            **config_kwargs,
        )
        self._autoscaler_v2 = autoscaler_v2

    def _generate_config(self, head_resources, worker_node_types, autoscaler_v2=False, **config_kwargs):
        example_yaml_path = os.path.join(os.path.dirname(__file__), "fake_cluster.yaml")
        with open(example_yaml_path) as f:
            base_config = yaml.safe_load(f)
        custom_config = copy.deepcopy(base_config)
        custom_config["available_node_types"] = worker_node_types
        custom_config["available_node_types"]["ray.head.default"] = {
            "resources": head_resources,
            "node_config": {},
            "max_workers": 0,
        }

        # Autoscaler v2 specific configs
        if autoscaler_v2:
            custom_config["provider"]["launch_multiple"] = True
            custom_config["provider"]["head_node_id"] = "fake_head_node_id"
        custom_config.update(config_kwargs)
        return custom_config

    def start(self, _system_config=None, override_env: dict | None = None):
        """Start the cluster.

        After this call returns, you can connect to the cluster with
        ray.init("auto").
        """
        subprocess.check_call(["ray", "stop", "--force"])
        _, fake_config = tempfile.mkstemp()
        with open(fake_config, "w") as f:
            f.write(json.dumps(self._config))
        cmd = [
            "ray",
            "start",
            f"--autoscaling-config={fake_config}",
            "--head",
        ]
        if "CPU" in self._head_resources:
            cmd.append("--num-cpus={}".format(self._head_resources.pop("CPU")))
        if "GPU" in self._head_resources:
            cmd.append("--num-gpus={}".format(self._head_resources.pop("GPU")))
        if "object_store_memory" in self._head_resources:
            cmd.append("--object-store-memory={}".format(self._head_resources.pop("object_store_memory")))
        if self._head_resources:
            cmd.append(f"--resources='{json.dumps(self._head_resources)}'")
        if _system_config is not None:
            cmd.append("--system-config={}".format(json.dumps(_system_config, separators=(",", ":"))))
        env = os.environ.copy()
        env.update({"AUTOSCALER_UPDATE_INTERVAL_S": "1", "RAY_FAKE_CLUSTER": "1"})
        if self._autoscaler_v2:
            # Set the necessary environment variables for autoscaler v2.
            env.update(
                {
                    "RAY_enable_autoscaler_v2": "1",
                    "RAY_CLOUD_INSTANCE_ID": "fake_head_node_id",
                    "RAY_OVERRIDE_NODE_ID_FOR_TESTING": "fake_head_node_id",
                }
            )
        if override_env:
            env.update(override_env)
        subprocess.check_call(cmd, env=env)

    def shutdown(self):
        """Terminate the cluster."""
        subprocess.check_call(["ray", "stop", "--force"])


@contextmanager
def autoscaling_cluster_context(head_resources: dict, worker_node_types: dict, autoscaler_v2: bool = False):
    """Context manager for managing AutoscalingCluster with proper setup and teardown.

    The cluster is created before ray.init() as requested, and properly cleaned up afterward.

    Note: Starting and stopping the cluster can be time consuming. Use sparingly, and reuse the cluster if possible.
    """
    try:
        from daft.daft import reset_runner

        reset_runner()

        if ray.is_initialized():
            ray.shutdown()

        # Create the autoscaling cluster before ray.init()
        cluster = AutoscalingCluster(
            head_resources=head_resources,
            worker_node_types=worker_node_types,
            autoscaler_v2=autoscaler_v2,
        )

        # Start the cluster
        cluster.start()

        # Initialize Ray with the cluster
        ray.init()

        yield cluster

    finally:
        # Clean shutdown sequence
        ray.shutdown()
        cluster.shutdown()

        reset_runner()



from __future__ import annotations

import io

import numpy as np
from PIL import Image

import daft
from daft import col
from daft.datatype import DataType
from daft.series import Series
from tests.cookbook.assets import ASSET_FOLDER


def test_image_resize_mixed_modes(with_morsel_size) -> None:
    rgba = np.ones((2, 2, 4), dtype=np.uint8)
    rgba[..., 1] = 2
    rgba[..., 2] = 3
    rgba[..., 3] = 4

    data = [
        rgba[..., :3],  # rgb
        rgba,  # RGBA
        np.arange(12, dtype=np.uint8).reshape((1, 4, 3)),  # RGB
        np.arange(12, dtype=np.uint8).reshape((3, 4)) * 10,  # L
        np.ones(24, dtype=np.uint8).reshape((3, 4, 2)) * 10,  # LA
        None,
    ]

    s = Series.from_pylist(data, dtype=DataType.python())
    df = daft.from_pydict({"img": s})

    target_dtype = DataType.image()
    df = df.select(df["img"].cast(target_dtype))

    assert df.schema()["img"].dtype == target_dtype

    df = df.with_column("resized", df["img"].image.resize(5, 5))

    assert df.schema()["resized"].dtype == target_dtype

    as_py = df.to_pydict()["resized"]

    first_resized = as_py[0]
    assert np.all(first_resized[..., 0] == 1)
    assert np.all(first_resized[..., 1] == 2)
    assert np.all(first_resized[..., 2] == 3)

    second_resized = as_py[1]
    assert np.all(second_resized[..., 0] == 1)
    assert np.all(second_resized[..., 1] == 2)
    assert np.all(second_resized[..., 2] == 3)
    assert np.all(second_resized[..., 3] == 4)

    for i in range(2, 4):
        resized_i = as_py[i]
        resized_i_gt = np.asarray(Image.fromarray(data[i]).resize((5, 5), resample=Image.BILINEAR)).reshape(5, 5, -1)
        assert np.all(resized_i == resized_i_gt), f"{i} does not match"

    # LA sampling doesn't work for some reason in PIL
    resized_i = as_py[4]
    assert np.all(resized_i == 10)

    assert as_py[-1] is None


def test_image_decode(with_morsel_size) -> None:
    df = (
        daft.from_glob_path(f"{ASSET_FOLDER}/images/**")
        .into_partitions(2)
        .with_column("image", col("path").url.download().image.decode().image.resize(10, 10))
    )
    target_dtype = DataType.image()
    assert df.schema()["image"].dtype == target_dtype
    df.collect()


def test_image_encode(with_morsel_size) -> None:
    file_format = "png"
    mode = "RGB"
    np_dtype = np.uint8
    shape = (4, 4, 3)
    arr = np.arange(np.prod(shape)).reshape(shape).astype(np_dtype)
    arrs = [arr, arr, arr]

    s = Series.from_pylist(arrs, dtype=DataType.python())

    df = daft.from_pydict({"img": s}).into_partitions(2)
    target_dtype = DataType.image(mode)
    df = df.select(df["img"].cast(target_dtype))
    assert df.schema()["img"].dtype == target_dtype

    df = df.with_column("encoded", df["img"].image.encode(file_format))
    assert df.schema()["encoded"].dtype == DataType.binary()

    pil_decoded_imgs = [np.asarray(Image.open(io.BytesIO(bytes_))) for bytes_ in df.to_pydict()["encoded"]]
    np.testing.assert_equal(pil_decoded_imgs, arrs)
